# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for gwpy.types.series."""

from __future__ import annotations

import warnings
from typing import (
    Generic,
    TypeVar,
    cast,
)

import numpy
import pytest
from astropy import units
from numpy.testing import assert_array_equal

from ...segments import Segment
from ...testing import utils
from .. import Array2D, Index, Series
from .test_array import TestArray as _TestArray

SeriesType = TypeVar("SeriesType", bound=Series)


class TestSeries(_TestArray[SeriesType], Generic[SeriesType]):
    """Test the `Series` class."""

    TEST_CLASS: type[SeriesType] = Series

    def assert_new(self, array: SeriesType):
        """Assert properties of a new `Series`."""
        super().assert_new(array)
        assert array.x0 == units.Quantity(0, self.TEST_CLASS._default_xunit)
        assert array.dx == units.Quantity(1, self.TEST_CLASS._default_xunit)

    # -- test properties -------------

    def test_x0(self, array: SeriesType):
        """Test `Series.x0`."""
        array.x0 = 5
        # test simple
        assert array.x0 == 5 * self.TEST_CLASS._default_xunit

        # test deleter
        del array.x0
        del array.x0
        assert array.x0 == 0 * self.TEST_CLASS._default_xunit

        # test quantity
        array.x0 = 5 * units.m
        assert array.x0 == units.Quantity(5, "m")

    def test_dx(self, array: SeriesType):
        """Test `Series.dx`."""
        array.dx = 5 * self.TEST_CLASS._default_xunit
        # test simple
        assert array.dx == units.Quantity(5, self.TEST_CLASS._default_xunit)

        # test deleter
        del array.dx
        del array.dx
        assert array.dx == units.Quantity(1, self.TEST_CLASS._default_xunit)

        # test quantity
        array.dx = 5 * units.m
        assert array.dx == units.Quantity(5, "m")

    def test_xindex(self):
        """Test `Series.xindex`."""
        x = numpy.linspace(0, 100, num=self.data.shape[0])

        # test simple
        series = self.create(xindex=x)
        utils.assert_quantity_equal(
            series.xindex,
            units.Quantity(x, self.TEST_CLASS._default_xunit),
        )

        # test deleter
        del series.xindex
        del series.xindex
        x1 = series.x0.value + series.shape[0] * series.dx.value
        x_default = numpy.linspace(
            series.x0.value,
            x1,
            num=series.shape[0],
            endpoint=False,
        )
        utils.assert_quantity_equal(
            series.xindex,
            units.Quantity(x_default, self.TEST_CLASS._default_xunit),
        )

        # test setting of x0 and dx
        series = self.create(xindex=units.Quantity(x, "Farad"))
        assert series.x0 == units.Quantity(x[0], "Farad")
        assert series.dx == units.Quantity(x[1] - x[0], "Farad")
        assert series.xunit == units.Farad
        assert series.xspan == (x[0], x[-1] + x[1] - x[0])

        # test that setting xindex warns about ignoring dx or x0
        with pytest.warns(
            UserWarning,
            match="xindex was given",
        ):
            series = self.create(xindex=units.Quantity(x, "Farad"), dx=1)
        with pytest.warns(
            UserWarning,
            match="xindex was given",
        ):
            series = self.create(xindex=units.Quantity(x, "Farad"), x0=0)

        # test non-regular xindex
        x = numpy.logspace(0, 2, num=self.data.shape[0])
        series = self.create(xindex=units.Quantity(x, "Mpc"))
        with pytest.raises(AttributeError):
            series.dx  # noqa: B018
        assert series.x0 == units.Quantity(1, "Mpc")
        assert series.xspan == (x[0], x[-1] + x[-1] - x[-2])

    def test_xindex_length_exception(self):
        """Test that xindex must match data length."""
        with pytest.raises(
            ValueError,
            match="xindex must have the same length as data",
        ):
            Series([1, 2, 3], xindex=[0])

    def test_xindex_dtype(self):
        """Test that xindex dtype matches x0 and dx."""
        x0 = numpy.longdouble(100)
        dx = numpy.float32(1e-4)
        series = self.create(x0=x0, dx=dx)
        assert series.xindex.dtype is x0.dtype

    def test_xunit(self):
        """Test `Series.xunit`."""
        unit = self.TEST_CLASS._default_xunit
        series = self.create(dx=4 * unit)
        assert series.xunit == unit
        assert series.x0 == 0 * unit
        assert series.dx == 4 * unit
        # for series only, test arbitrary xunit
        if self.TEST_CLASS is (Series, Array2D):
            series = self.create(dx=4, xunit=units.m)
            assert series.x0 == 0 * units.m
            assert series.dx == 4 * units.m

    def test_xspan(self):
        """Test `Series.xspan`."""
        # test normal
        series = self.create(x0=1, dx=1)
        assert series.xspan == (1, 1 + 1 * series.shape[0])
        assert isinstance(series.xspan, Segment)
        # test from irregular xindex
        x = numpy.logspace(0, 2, num=self.data.shape[0])
        series = self.create(xindex=x)
        assert series.xspan == (x[0], x[-1] + x[-1] - x[-2])

    # -- test i/o --------------------

    def test_read_write_csv(self, array: SeriesType):
        """Test reading and writing a `Series` to CSV."""
        utils.test_read_write(
            array,
            "csv",
            assert_equal=utils.assert_quantity_sub_equal,
            assert_kw={
                "exclude": [
                    "name",
                    "channel",
                    "unit",
                ],
            },
        )

    # -- test methods ----------------

    def test_getitem(self, array: SeriesType):
        """Test `Series[...]` item access."""
        # item access
        utils.assert_quantity_equal(
            array[0],
            units.Quantity(array.value[0], array.unit),
        )

        # slice
        utils.assert_quantity_equal(
            array[1::2],
            self.TEST_CLASS(
                array.value[1::2],
                x0=array.x0 + array.dx,
                dx=array.dx * 2,
                name=array.name,
                epoch=array.epoch,
                unit=array.unit,
            ),
        )

        # index array
        a = numpy.array([3, 4, 1, 2])
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message="xindex was given",
                category=UserWarning,
            )
            utils.assert_quantity_equal(
                array[a],
                self.TEST_CLASS(
                    array.value[a],
                    xindex=array.xindex[a],
                    name=array.name,
                    epoch=array.epoch,
                    unit=array.unit,
                ),
            )

    def test_getitem_index(self, array: SeriesType):
        """Test that __getitem__ also applies to an xindex.

        When subsetting a Series with an iterable of integer indices,
        make sure that the xindex, if it exists, is also subsetted. Tests
        regression against https://gitlab.com/gwpy/gwpy/-/issues/1680.
        """
        array.xindex  # create xindex  # noqa: B018
        indices = numpy.array([0, 1, len(array) - 1])
        newarray = array[indices]

        assert len(newarray) == 3
        assert len(newarray) == len(newarray.value)
        assert len(newarray.value) == len(newarray.xindex)

    def test_getitem_list_index(self, array: SeriesType):
        """Test that __getitem__ works with list and numpy.array."""
        indices = numpy.array([0, 1, len(array) - 1])
        lindices = [0, 1, len(array) - 1]
        utils.assert_quantity_sub_equal(array[indices], array[lindices])

    def test_single_getitem_not_created(self, array: SeriesType):
        """Test that array[i] does not return an object with a new _xindex."""
        # check that there is no xindex when a single value is accessed
        with pytest.raises(
            AttributeError,
            match=r"'.*' object has no",
        ):
            array[0].xindex  # noqa: B018

        # we don't need this, we don't want it accidentally injected
        with pytest.raises(
            AttributeError,
            match=r"'.*' object has no",
        ):
            array[0]._xindex  # noqa: B018

    def test_empty_slice(self, array: SeriesType):
        """Check that we can slice a `Series` into nothing.

        This tests against a bug in sliceutils.py.
        """
        a2 = array[:0]
        assert a2.x0 == array.x0
        assert a2.dx == array.dx

        idx = numpy.array([False] * array.shape[0])  # False slice array
        a3 = array[idx]
        utils.assert_quantity_sub_equal(a2, a3)

        a4 = array[idx[:0]]  # empty slice array
        utils.assert_quantity_sub_equal(a2, a4)

    def test_zip(self, array: SeriesType):
        """Test `Series.zip`."""
        z = array.zip()
        assert_array_equal(
            z,
            numpy.column_stack((array.xindex.value, array.value)),
        )

    def test_crop(self, array: SeriesType):
        """Test basic functionality of `Series.crop`."""
        # all defaults
        utils.assert_quantity_equal(array, array.crop())

        # normal operation
        a2 = array.crop(10, 20)
        utils.assert_quantity_equal(array[10:20], a2)

    def test_crop_warnings(self, array: SeriesType):
        """Test that `Series.crop` emits warnings when it is supposed to."""
        with pytest.warns(
            UserWarning,
            match="crop given start",
        ):
            array.crop(array.xspan[0] - 1, array.xspan[1])
        with pytest.warns(
            UserWarning,
            match="crop given end",
        ):
            array.crop(array.xspan[0], array.xspan[1] + 1)

    def test_crop_irregular(self):
        """Test `Series.crop` with an irregular index."""
        x = numpy.linspace(0, 100, num=self.data.shape[0])

        # add shift to second half of times to create irregular space
        x[x > 50] += 1

        series = self.create(xindex=x)

        cropped = series.crop(end=75)
        utils.assert_quantity_equal(series[x < 75], cropped)

        cropped = series.crop(start=25)
        utils.assert_quantity_equal(series[x > 25], cropped)

        cropped = series.crop(start=25, end=75)
        utils.assert_quantity_equal(series[(x > 25) & (x < 75)], cropped)

    def test_crop_float_precision_last_value(self):
        """Verify the float precision of Series.crop given the last index.

        This tests against regression of
        https://gitlab.com/gwpy/gwpy/-/issues/1601.
        """
        # construct empty data array with the right shape for this array object
        shape = (101,) * self.TEST_CLASS._ndim
        series = self.TEST_CLASS(numpy.empty(shape), dx=0.01)

        # assert that when we crop it, we only crop a single sample
        cropped = series.crop(end=1.0)
        utils.assert_quantity_equal(series[:-1], cropped)

    def test_crop_float_precision_last_value_float(self):
        """Verify the float precision of the crop function with float end.

        This tests against regression of
        https://gitlab.com/gwpy/gwpy/-/issues/1656.
        """
        arrlen = 500
        xmax = 0.508463154883984
        x_series = numpy.linspace(0, xmax, arrlen)
        series = Series([0] * arrlen, xindex=x_series)
        expected = series.xindex[-2]
        assert series.crop(end=xmax).xindex[-1] == expected

    def test_crop_between_grid_points_is_floored(self):
        """Test that when we crop between xindex values, the result is floored.

        This tests against regression of
        https://gitlab.com/gwpy/gwpy/-/issues/1656.
        """
        # e.g. x = [1, 2, 3], end = 2.5, result = [1, 2]
        series = Series([0] * 3, xindex=[1, 2, 3])
        assert_array_equal(series.crop(end=2.5).xindex, [1, 2])

        series = Series([0] * 3, xindex=[1, 2, 3])
        assert_array_equal(series.crop(start=2.5).xindex, [3])

        series = Series([0] * 5, xindex=[1, 2, 3, 4, 5])
        assert_array_equal(series.crop(start=2.5, end=4.5).xindex, [3, 4])

    def test_crop_float_precision_near_last_value_float(self):
        """Test the float precision of Series.crop with arg just under end.

        This tests against regression of
        https://gitlab.com/gwpy/gwpy/-/issues/1656.
        """
        arrlen = 500
        xmax = 0.508463154883984
        x_series = numpy.linspace(0, xmax, arrlen)
        series = Series([0] * arrlen, xindex=x_series)
        expected = series.xindex[-2]
        mid = 0.6 * series.xindex[-2] + 0.4 * series.xindex[-1]
        assert series.crop(end=mid).xindex[-1] == expected

    def test_crop_float_precision_first_value_float(self):
        """Verify the float precision of the crop function with float start.

        This tests against regression of
        https://gitlab.com/gwpy/gwpy/-/issues/1656.
        """
        arrlen = 500
        xmin = 0.508463154883984
        x_series = numpy.linspace(xmin, 1.0, arrlen)
        series = Series([0] * arrlen, xindex=x_series)
        expected = series.xindex[0]
        assert series.crop(start=xmin).xindex[0] == expected

    def test_is_compatible(self, array: SeriesType):
        """Test the `Series.is_compatible` method."""
        other = self.create(name="TEST CASE 2")
        assert array.is_compatible(other)

    def test_is_compatible_error_dx(self, array: SeriesType):
        """Check that `Series.is_compatible` errors with mismatching ``dx``."""
        other = self.create(dx=2)
        with pytest.raises(ValueError, match="sample sizes do not match"):
            array.is_compatible(other)

    def test_is_compatible_error_unit(self, array: SeriesType):
        """Check that `Series.is_compatible` errors with mismatching ``unit``."""
        other = self.create(unit="m")
        with pytest.raises(ValueError, match="units do not match"):
            array.is_compatible(other)

    def test_is_compatible_xindex(self):
        """Check that irregular arrays are compatible if their xindexes match."""
        x = numpy.logspace(0, 2, num=self.data.shape[0])
        a = self.create(xindex=x)
        b = self.create(xindex=x)
        assert a.is_compatible(b)

    def test_is_compatible_error_xindex(self, array: SeriesType):
        """Check that `Series.is_compatible` errors with mismatching indexes."""
        x = numpy.logspace(0, 2, num=self.data.shape[0])
        other = self.create(xindex=x)
        with pytest.raises(ValueError, match="indexes do not match"):
            array.is_compatible(other)

    def test_is_contiguous(self, array: SeriesType):
        """Test `Series.is_contiguous`."""
        a2 = self.create(x0=array.xspan[1])
        assert array.is_contiguous(a2) == 1
        assert array.is_contiguous(a2.value) == 1

        ts3 = self.create(x0=array.xspan[1] + 1)
        assert array.is_contiguous(ts3) == 0

        ts4 = self.create(x0=-array.xspan[1])
        assert array.is_contiguous(ts4) == -1

    def test_append(self, array: SeriesType):
        """Test `Series.append`."""
        a2 = self.create(x0=array.xspan[1])

        # test basic append
        a3 = array.append(a2, inplace=False)
        assert a3.epoch == array.epoch
        assert a3.x0 == array.x0
        assert a3.size == array.size + a2.size
        assert a3.xspan == array.xspan + a2.xspan
        assert_array_equal(a3.value[: array.shape[0]], array.value)
        assert_array_equal(a3.value[-a2.shape[0] :], a2.value)

        # check that appending again causes a problem
        with pytest.raises(
            ValueError,
            match=f"Cannot append discontiguous {type(a3).__name__}",
        ):
            a3.append(array)

        # test appending with one xindex deletes it in the output
        array.xindex  # noqa: B018
        a3 = array.append(a2, inplace=False)
        assert hasattr(a3, "_xindex") is False

        # test appending with both xindex appends as well
        array.xindex  # noqa: B018
        a2.xindex  # noqa: B018
        a3 = array.append(a2, inplace=False)
        assert hasattr(a3, "_xindex")
        assert_array_equal(
            a3.xindex.value,
            numpy.concatenate((array.xindex.value, a2.xindex.value)),
        )

        # test appending with one only and not resize
        del a2.xindex
        a3 = array.append(a2, inplace=False, resize=False)
        assert a3.x0 == array.x0 + array.dx * a2.shape[0]

        # test discontiguous appends - gap='raise'
        a3 = self.create(x0=array.xspan[1] + 1)
        ts4 = array.copy()
        with pytest.raises(
            ValueError,
            match=f"Cannot append discontiguous {type(array).__name__}",
        ):
            array.append(a3)

        # test gap='ignore'
        ts4.append(a3, gap="ignore")
        assert ts4.shape[0] == array.shape[0] + a3.shape[0]
        assert_array_equal(ts4.value, numpy.concatenate((array.value, a3.value)))

        # test gap='pad'
        ts4 = array.copy()
        ts4.append(a3, gap="pad", pad=0)
        assert ts4.shape[0] == array.shape[0] + 1 + a3.shape[0]
        z = numpy.zeros((1, *array.shape[1:]))
        assert_array_equal(
            ts4.value,
            numpy.concatenate((array.value, z, a3.value)),
        )

    def test_prepend(self, array: SeriesType):
        """Test `Series.prepend`."""
        a2 = cast("SeriesType", self.create(x0=array.xspan[1]) * 2)
        a3 = a2.prepend(array, inplace=False)
        assert a3.x0 == array.x0
        assert a3.size == array.size + a2.size
        assert a3.xspan == array.xspan + a2.xspan
        with pytest.raises(
            ValueError,
            match=f"Cannot append discontiguous {type(a3).__name__}",
        ):
            a3.prepend(array)
        assert_array_equal(a3.value[: array.shape[0]], array.value)
        assert_array_equal(a3.value[-a2.shape[0] :], a2.value)

    def test_update(self):
        """Test `Series.update`."""
        ts1 = self.create()
        ts2 = self.create(x0=ts1.xspan[1])[: ts1.size // 2]
        ts3 = ts1.update(ts2, inplace=False)
        assert ts3.x0 == ts1.x0 + abs(ts2.xspan) * ts1.x0.unit
        assert ts3.size == ts1.size
        with pytest.raises(
            ValueError,
            match=f"Cannot append discontiguous {type(ts3).__name__}",
        ):
            ts3.update(ts1)

    def test_pad(self):
        """Test `Series.pad`."""
        ts1 = self.create()
        ts2 = ts1.pad(10)
        assert ts2.shape[0] == ts1.shape[0] + 20
        assert_array_equal(
            ts2.value,
            numpy.concatenate((numpy.zeros(10), ts1.value, numpy.zeros(10))),
        )
        assert ts2.x0 == ts1.x0 - 10 * ts1.x0.unit

    def test_pad_index(self):
        """Check that `Series.pad` correctly returns a padded index array."""
        ts1 = self.create()
        ts1.xindex  # <- create the index for ts1  # noqa: B018

        ts2 = ts1.pad(10)

        start: float
        end: float
        start, end = ts2.xspan
        assert_array_equal(
            ts2.xindex,
            Index(
                numpy.linspace(start, end, num=ts2.shape[0], endpoint=False),
                unit=ts1.xindex.unit,
            ),
        )

    def test_pad_asymmetric(self):
        """Test `Series.pad` with asymmetric padding."""
        ts1 = self.create()
        ts2 = ts1.pad((20, 10))
        assert ts2.shape[0] == ts1.shape[0] + 30
        assert_array_equal(
            ts2.value,
            numpy.concatenate((numpy.zeros(20), ts1.value, numpy.zeros(10))),
        )
        assert ts2.x0 == ts1.x0 - 20 * ts1.x0.unit

    def test_diff(self, array: SeriesType):
        """Test `Series.diff`."""
        """Test the `Series.diff` method.

        This just ensures that the returned `Series` has the right length
        and the right x0
        """
        diff = array.diff(axis=0)
        assert isinstance(diff, type(array))
        assert array.shape[0] - 1 == diff.shape[0]
        assert diff.x0 == array.x0 + array.dx
        assert diff.xspan[1] == array.xspan[1]
        assert diff.channel == array.channel
        # test n=3
        diff = array.diff(n=3)
        assert array.shape[-1] - 3 == diff.shape[-1]
        assert diff.x0 == array.x0 + array.dx * 3

    def test_value_at(self):
        """Test `Series.value_at`."""
        ts1 = self.create(dx=0.5)
        assert ts1.value_at(1.5) == self.data[3] * ts1.unit
        assert ts1.value_at(1.5 * ts1.xunit) == self.data[3] * ts1.unit
        with pytest.raises(IndexError):
            ts1.value_at(1.6)
        # test TimeSeries unit conversion
        if ts1.xunit == units.s:
            assert ts1.value_at(1500 * units.millisecond) == (self.data[3] * ts1.unit)
        # test FrequencySeries unit conversion
        elif ts1.xunit == units.Hz:
            assert ts1.value_at(1500 * units.milliHertz) == (self.data[3] * ts1.unit)

    def test_shift(self):
        """Test `Series.shift`."""
        a = self.create(x0=0, dx=1, xunit="s")
        x0 = a.x0.copy()
        a.shift(5)
        assert a.x0 == x0 + 5 * x0.unit

        a.shift("1 hour")
        assert a.x0 == x0 + 3605 * x0.unit

        a.shift(-0.007)
        assert a.x0 == x0 + (3604.993) * x0.unit

        with pytest.raises(
            ValueError,
            match=r"'Hz' \(frequency\) and 's' \(time\) are not convertible",
        ):
            a.shift("1 Hz")

    @pytest.mark.parametrize("inplace", [True, False])
    def test_inject(self, inplace: bool):
        """Test `Series.inject()`."""
        # Create a series of zeros (supporting multi-dim series)
        duration = 16
        ndim = self.TEST_CLASS._ndim
        nsamp = duration * ndim
        shape = () if ndim == 1 else (ndim,)
        zshape = (duration, *shape)
        data = self.TEST_CLASS(numpy.zeros(zshape))

        # Create a second series to inject into the first
        ishape = (duration // 2, *shape)
        start = duration // 4
        injection = self.TEST_CLASS(
            numpy.linspace(1, 2, num=nsamp // 2).reshape(ishape),
            x0=start + .1,  # non-grid-aligned start time
        )

        # Test that we recover this injection when we add it to data
        with pytest.warns(
            UserWarning,
            match="will round",
        ):
            new = data.inject(injection, inplace=inplace)
        assert new.unit == data.unit
        assert new.size == data.size

        # Check that the injection is where we expect it to be
        nzind = new.value.nonzero()
        assert nzind[0].size == injection.size
        first = tuple(nzind[i][0] for i in range(len(nzind)))
        assert first[0] == start * data.dx.value

        # Check that the injected values are correct
        utils.assert_allclose(new.value[nzind], injection.value.flatten())

        # Check that inplace does what we expect
        if inplace:
            assert new is data
        else:
            utils.assert_allclose(data.value, numpy.zeros(zshape))



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Read gravitational-wave frame (GWF) files using the LALFrame API.

The frame format is defined in :dcc:`LIGO-T970130`.
"""

from __future__ import annotations

import contextlib
import os.path
import warnings
from typing import TYPE_CHECKING

import lal
import lalframe

from ....io.gwf import lalframe as io_gwf_lalframe
from ....io.utils import (
    file_list,
    file_path,
)
from ....segments import Segment
from ....utils import lal as lalutils
from ... import TimeSeries

if TYPE_CHECKING:
    from pathlib import Path
    from typing import IO

    from ....time import LIGOTimeGPS
    from ... import (
        TimeSeriesBase,
        TimeSeriesBaseDict,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


# -- utilities -----------------------

def open_data_source(
    source: str | Path | IO | lal.Cache | list[str | Path | IO],
) -> lalframe.FrStream:
    """Open a GWF file source into a |lalframe.FrStream|_ object.

    Parameters
    ----------
    source : `str`, `file`, `list`
        Data source to read.

    Returns
    -------
    stream : |lalframe.FrStream|_
        An open `lalframe.FrStream`.

    Raises
    ------
    ValueError
        If the input format cannot be identified.
    """
    # -- preformatting

    with contextlib.suppress(ValueError):
        source = file_path(source)

    # import cache from file
    if (
        isinstance(source, str)
        and source.endswith((".lcf", ".cache"))
    ):
        source = lal.CacheImport(source)

    # reformat cache (or any list of files) as a lal cache object
    if isinstance(source, list):
        cache = lal.Cache()
        for entry in file_list(source):
            cache = lal.CacheMerge(cache, lal.CacheGlob(*os.path.split(entry)))
        source = cache

    # -- now we have a lal.Cache or a filename

    # read lal cache object
    if isinstance(source, lal.Cache):
        return lalframe.FrStreamCacheOpen(source)

    # read single file
    if isinstance(source, str):
        return lalframe.FrStreamOpen(*map(str, os.path.split(source)))

    msg = f"Don't know how to open data source of type '{type(source).__name__}'"
    raise ValueError(msg)


def get_stream_segment(stream: lalframe.FrStream) -> Segment:
    """Get the GPS end time of the last frame in this stream.

    Parameters
    ----------
    stream : |lalframe.FrStream|_
        Stream of data to search.

    Returns
    -------
    segment: `gwpy.segments.Segment`
        The GPS ``[start, end)`` segment of data covered by this stream.
    """
    epoch = lal.LIGOTimeGPS(stream.epoch)
    try:
        # rewind stream to the start
        lalframe.FrStreamRewind(stream)
        start = lal.LIGOTimeGPS(stream.epoch)
        end = 0.
        # loop over each file in the stream cache and query its duration
        for _ in range(stream.cache.length):
            for j in range(lalframe.FrFileQueryNFrame(stream.file)):
                # get GPS start and duration of this frame
                fstart = lalframe.FrFileQueryGTime(
                    lal.LIGOTimeGPS(),
                    stream.file,
                    j,
                )
                dt = lalframe.FrFileQueryDt(stream.file, j)
                # record the end time of the last frame
                end = max(end, fstart + dt)
                # move on to the next frame
                lalframe.FrStreamNext(stream)
        return Segment(start, end)
    finally:
        # reset stream to where we started and return
        lalframe.FrStreamSeek(stream, epoch)


def get_stream_duration(stream: lalframe.FrStream) -> float:
    """Calculate the duration of time stored in a frame stream.

    Parameters
    ----------
    stream : |lalframe.FrStream|_
        Stream of data to search.

    Returns
    -------
    duration : `float`
        The duration (seconds) of the data for this channel.
    """
    return abs(get_stream_segment(stream))


# -- read ----------------------------

def read(
    source: str | Path | IO | list[str | Path | IO],
    channels: list[str],
    start: LIGOTimeGPS | None = None,
    end: LIGOTimeGPS | None = None,
    type: str | dict[str, str] | None = None,
    series_class: type[TimeSeriesBase] = TimeSeries,
    scaled: bool | None = None,
) -> TimeSeriesBaseDict:
    """Read data from one or more GWF files using the LALFrame API."""
    # scaled must be provided to provide a consistent API with frameCPP
    if scaled is not None:
        warnings.warn(
            "the `scaled` keyword argument is not supported by lalframe, "
            "if you require ADC scaling, please install "
            "python-ldas-tools-framecpp and use `backend='frameCPP'` when "
            "reading GWF data",
            stacklevel=2,
        )

    stream = open_data_source(source)

    # parse times and restrict to available data
    streamstart, streamend = get_stream_segment(stream)
    if start is None:
        startgps = streamstart
    else:
        startgps = lalutils.to_lal_ligotimegps(start)
    if startgps >= streamend:
        msg = f"cannot read data starting at {startgps}, stream ends at {streamend}"
        raise ValueError(msg)
    startgps = max(streamstart, startgps)
    if end is None:
        endgps = streamend
    else:
        endgps = lalutils.to_lal_ligotimegps(end)
    if endgps <= streamstart:
        msg = f"cannot read data ending at {endgps}, stream starts at {streamstart}"
        raise ValueError(msg)
    endgps = min(streamend, endgps)
    span = Segment(startgps, endgps)
    duration = float(abs(span))

    if endgps <= startgps:
        msg = "cannot read data with non-positive duration"
        raise ValueError(msg)

    # read data
    out = series_class.DictClass()
    for name in channels:
        lalframe.FrStreamSeek(stream, startgps)
        # read this channel and convert to GWpy struct
        ts = _read_channel(stream, str(name), start=startgps, duration=duration)
        out[name] = series_class.from_lal(ts, copy=False)
        # check what we got back
        tsend = ts.epoch + getattr(ts.data.data, "size", 0) * ts.deltaT
        if (missing := endgps - tsend) >= ts.deltaT:  # one sample missing
            # if the available data simply didn't go all the way to the end
            # lalframe would have errored (and we protect against that above),
            # so we know there is data missing.
            # see https://git.ligo.org/lscsoft/lalsuite/-/issues/710
            ts = _read_channel(
                stream,
                str(name),
                start=tsend,
                duration=float(missing),
            )
            out[name] = out[name].append(ts.data.data, inplace=False)

    return out


def _read_channel(
    stream: lalframe.FrStream,
    channel: str,
    start: lal.LIGOTimeGPS,
    duration: float,
) -> lalutils.LALTimeSeriesType:
    """Read a channel from this stream."""
    # find the data type for this channel
    try:
        dtype = lalframe.FrStreamGetTimeSeriesType(channel, stream)
    except RuntimeError as exc:
        if str(exc).lower() == "wrong name":
            msg = f"channel '{channel}' not found"
            raise ValueError(msg) from exc
        raise
    # get the appropriate reader for this type
    reader = lalutils.find_typed_function(
        dtype,
        "FrStreamRead",
        "TimeSeries",
        module=lalframe,
    )
    # read the data
    return reader(stream, channel, start, duration, 0)


# -- write ---------------------------

def write(
    tsdict: TimeSeriesBaseDict,
    outfile: str | Path | IO,
    start: LIGOTimeGPS,
    end: LIGOTimeGPS,
    type: str | None = None,
    name: str | None = None,
    run: int = 0,
    compression: int | str | None = None,
    compression_level: int | None = None,
) -> None:
    """Write data to a GWF file using the LALFrame API."""
    startgps = lalutils.to_lal_ligotimegps(start)
    endgps = lalutils.to_lal_ligotimegps(end)

    ifos = {
        ts.channel.ifo for ts in tsdict.values() if (
            ts.channel
            and ts.channel.ifo
        )
    }

    # LALFrame Python bindings (as we use them) don't allow specifying
    # compression options
    if compression is not None or compression_level is not None:
        warnings.warn(
            "LALFrame backend does not support compression options, "
            "compression and compression_level values will be ignored",
            stacklevel=2,
        )


    # create new frame
    frame = io_gwf_lalframe.create_frame(
        startgps,
        float(endgps - startgps),
        name or "gwpy",
        run,
        0,
        ifos,
    )

    # append each series
    for series in tsdict.values():
        # get type
        ctype = (
            type
            or getattr(series.channel, "_ctype", "proc")
            or "proc"
        ).title()

        # convert to LAL
        lalseries = series.to_lal()

        # find adder
        add_ = lalutils.find_typed_function(
            series.dtype,
            "FrameAdd",
            f"TimeSeries{ctype}Data",
            module=lalframe,
        )

        # add time series to frame
        add_(frame, lalseries)

    # write frame
    io_gwf_lalframe.write_frames(outfile, [frame])



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for :mod:`gwpy.astro.range`."""

from __future__ import annotations

from unittest import mock

import pytest
from astropy import units
from scipy.integrate import trapezoid

from ... import astro
from ...frequencyseries import FrequencySeries
from ...spectrogram import Spectrogram
from ...testing import utils
from ...timeseries import TimeSeries

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
__credits__ = "Alex Urban <alexander.urban@ligo.org>"

TEST_RESULTS = {
    "sensemon_range": 19.332958991178117 * units.Mpc,
    "inspiral_range": 18.519899937121536 * units.Mpc,
    "inspiral_range_snr_10": 14.82352747 * units.Mpc,
    "burst_range": 13.592140825568954 * units.Mpc,
}


@pytest.fixture(scope="module")
def hoft() -> TimeSeries:
    """Read the LIGO-Livingston strain data from the test HDF5 file."""
    return TimeSeries.read(
        utils.TEST_HDF5_FILE,
        "L1:LDAS-STRAIN",
        format="hdf5",
    )


@pytest.fixture(scope="module")
def psd(hoft) -> FrequencySeries:
    """Return the PSD for ``hoft``."""
    return hoft.psd(
        .4,
        overlap=.2,
        method="median",
        window=("kaiser", 24),
    )


# -- sensemon ------------------------

def test_sensemon_range_psd(psd):
    """Test :func:`gwpy.astro.sensemon_range_psd`."""
    fisco = astro.range._get_isco_frequency(1.4, 1.4).value
    frange = (psd.frequencies.value < fisco)
    r = astro.sensemon_range_psd(psd[frange])
    assert isinstance(r, FrequencySeries)
    utils.assert_quantity_almost_equal(
        trapezoid(r, r.frequencies) ** (1 / 2.),
        TEST_RESULTS["sensemon_range"],
    )
    assert r.f0.value > 0


def test_sensemon_range(psd):
    """Test :func:`gwpy.astro.sensemon_range`."""
    r = astro.sensemon_range(psd)
    utils.assert_quantity_almost_equal(r, TEST_RESULTS["sensemon_range"])


# -- inspiral-range ------------------

@mock.patch.dict("sys.modules", {"inspiral_range": None})
def test_inspiral_range_missing_dep(psd):
    """Test that :func:`gwpy.astro.inspiral_range` handles a failed import."""
    with pytest.raises(
        ImportError,
        match="extra package 'inspiral-range'",
    ):
        astro.inspiral_range(psd)


@pytest.mark.requires("inspiral_range")
def test_inspiral_range_psd(psd):
    """Test :func:`gwpy.astro.inspiral_range_psd`."""
    frange = (psd.frequencies.value < 4096)
    r = astro.inspiral_range_psd(psd[frange])
    assert isinstance(r, FrequencySeries)
    utils.assert_quantity_almost_equal(
        trapezoid(r, r.frequencies) ** (1 / 2.),
        TEST_RESULTS["inspiral_range"],
    )
    assert r.f0.value > 0


@pytest.mark.requires("inspiral_range")
@pytest.mark.parametrize(("snr", "expected"), [
    (8, TEST_RESULTS["inspiral_range"]),
    (10, TEST_RESULTS["inspiral_range_snr_10"]),
])
def test_inspiral_range(psd, snr, expected):
    """Test :func:`gwpy.astro.inspiral_range`."""
    result = astro.inspiral_range(psd, snr=snr)
    utils.assert_quantity_almost_equal(result, expected)


# -- burst range ---------------------

def test_burst_range_spectrum(psd):
    """Test :func:`gwpy.astro.burst_range_spectrum`."""
    f = psd.frequencies
    frange = (f.value >= 100) & (f.value < 500)
    r = astro.burst_range_spectrum(psd[frange])
    assert isinstance(r, FrequencySeries)
    utils.assert_quantity_almost_equal(
        (trapezoid(r ** 3, f[frange]) / (400 * units.Hz)) ** (1 / 3.),
        TEST_RESULTS["burst_range"],
    )
    assert r.f0.value > 0


def test_burst_range(psd):
    """Test :func:`gwpy.astro.burst_range`."""
    r = astro.burst_range(psd)
    utils.assert_quantity_almost_equal(r, TEST_RESULTS["burst_range"])


# -- timeseries/spectrogram wrappers -

@pytest.mark.parametrize("rangekwargs", [
    pytest.param(
        {"mass1": 1.4, "mass2": 1.4},
        marks=[pytest.mark.requires("inspiral_range")],
        id="inspiral_range",
    ),
    pytest.param(
        {"mass1": 1.4, "mass2": 1.4, "range_func": astro.sensemon_range},
        id="sensemon_range",
    ),
    pytest.param({"energy": 1e-2}, id="burst_range"),
])
def test_range_timeseries(hoft, rangekwargs):
    """Test :func:`gwpy.astro.range_timeseries` works in simple conditions."""
    trends = astro.range_timeseries(
        hoft,
        0.5,
        fftlength=0.25,
        overlap=0.125,
        method="median",
        nproc=2,
        **rangekwargs,
    )
    assert isinstance(trends, TimeSeries)
    assert trends.size == 2
    assert trends.unit == "Mpc"
    assert trends.dt == 0.5 * units.second


@pytest.mark.parametrize(("rangekwargs", "outunit"), [
    pytest.param(
        {"mass1": 1.4, "mass2": 1.4},
        units.Mpc ** 2 / units.Hz,
        marks=[pytest.mark.requires("inspiral_range")],
        id="inspiral_range",
    ),
    pytest.param(
        {"mass1": 1.4, "mass2": 1.4, "range_func": astro.sensemon_range_psd},
        units.Mpc ** 2 / units.Hz,
        id="sensemon_range",
    ),
    pytest.param({"energy": 1e-2}, units.Mpc, id="burst_range"),
])
def test_range_spectrogram(hoft, rangekwargs, outunit):
    """Test :func:`gwpy.astro.range_spectrogram` works in simple conditions."""
    spec = astro.range_spectrogram(
        hoft,
        0.5,
        fftlength=0.25,
        overlap=0.125,
        method="median",
        nproc=2,
        **rangekwargs,
    )
    assert isinstance(spec, Spectrogram)
    assert spec.shape[0] == 2
    assert spec.unit == outunit
    assert spec.f0 == spec.df
    assert spec.dt == 0.5 * units.second
    assert spec.df == 4 * units.Hertz


@pytest.mark.parametrize("range_func", [
    astro.range_timeseries,
    astro.range_spectrogram,
])
def test_range_incompatible_input(range_func):
    """Test that ``range_func`` handles incompatible inputs appropriately."""
    with pytest.raises(
        TypeError,
        match="Could not produce a spectrogram from the input",
    ):
        range_func(2, 0.5)



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.pycbc` and its integration with `EventTable`."""

import h5py
import numpy
import pytest
from numpy.random import default_rng

from ...frequencyseries import FrequencySeries
from ...testing.utils import (
    assert_array_equal,
    assert_quantity_sub_equal,
    assert_table_equal,
)
from .. import EventTable
from ..filter import filter_table
from ..io import pycbc as io_pycbc
from .utils import random_table

RNG = default_rng()

# -- fixtures ------------------------

@pytest.fixture
def h5file():
    """Create an empty in-memory HDF5 file."""
    with h5py.File(
        "test",
        mode="w-",
        driver="core",
        backing_store=False,
    ) as h5f:
        yield h5f


@pytest.fixture
def pycbclivetable():
    """Create an `EventTable` in PyCBC format with 100 rows."""
    return random_table(
        names = [
            "a",
            "b",
            "c",
            "chisq",
            "chisq_dof",
            "d",
            "e",
            "f",
            "mass1",
            "mass2",
            "snr",
        ],
        length=100,
    )


@pytest.fixture
def pycbclivepsd():
    """Create a mock PSD."""
    return FrequencySeries(RNG.normal(size=1000), df=1)


@pytest.fixture
def pycbclivefile(tmp_path, pycbclivetable, pycbclivepsd):
    """Create a fully-formed PyCBC-format HDF5 file."""
    # create table
    loudest = (pycbclivetable["snr"] > 500).nonzero()[0]

    # manually create pycbc_live-format HDF5 file
    tmp = tmp_path / "X1-Live-0-0.hdf"
    with h5py.File(tmp, "w") as h5f:
        group = h5f.create_group("X1")
        for col in pycbclivetable.columns:
            group.create_dataset(data=pycbclivetable[col], name=col)
        group.create_dataset("loudest", data=loudest)
        group.create_dataset("psd", data=pycbclivepsd.value)
        group["psd"].attrs["delta_f"] = pycbclivepsd.df.to("Hz").value

    return tmp


# -- internal tests ------------------

@pytest.mark.parametrize(("filename", "result"), [
    ("X1-Live-0-0.h5", True),
    ("X1-Live-0-0.hdf5", True),
    ("X1-MY_DATA-0-0.h5", False),
])
def test_identify_pycbc_live(tmp_path, filename, result):
    """Test :func:`identify_pycbc_live`."""
    path = tmp_path / filename
    h5py.File(path, "w").close()
    assert io_pycbc.identify_pycbc_live("read", path, None) is result
    with path.open("rb") as h5f:
        assert io_pycbc.identify_pycbc_live("read", path, h5f) is result


def test_empty_hdf5_file(h5file):
    """Test `empty_hdf5_file`."""
    assert io_pycbc.empty_hdf5_file(h5file)


def test_empty_hdf5_file_group(h5file):
    """Test `empty_hdf5_file` with a group."""
    h5file.create_group("H1")
    assert io_pycbc.empty_hdf5_file(h5file)
    assert io_pycbc.empty_hdf5_file(h5file, ifo="H1")


def test_empty_hdf5_file_datasets(h5file):
    """Test `empty_hdf5_file` with (empty) datasets."""
    h1group = h5file.create_group("H1")
    h1group.create_dataset("psd", data=numpy.empty(10))
    assert io_pycbc.empty_hdf5_file(h5file, ifo="H1")

    h1group.create_dataset("gates", data=numpy.empty(10))
    assert io_pycbc.empty_hdf5_file(h5file, ifo="H1")

    h1group.create_dataset("time", data=numpy.empty(10))
    assert not io_pycbc.empty_hdf5_file(h5file, ifo="H1")


# -- EventTable integration tests ----

@pytest.mark.parametrize("fmt", [
    None,  # should default to hdf5.pycbc_live as the highest-priority format
    "hdf5.pycbc_live",
])
def test_read_pycbc_live(pycbclivetable, pycbclivefile, fmt):
    """Test reading a PyCBC-Live file into an `EventTable`."""
    table = EventTable.read(pycbclivefile, format=fmt)
    assert_table_equal(pycbclivetable, table)
    assert table.meta["ifo"] == "X1"


def test_read_pycbc_live_kwargs(pycbclivetable, pycbclivefile):
    """Test reading a PyCBC-Live file into an `EventTable` with kwargs."""
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        ifo="X1",
    )
    assert_table_equal(pycbclivetable, table)


def test_read_pycbc_live_loudest(pycbclivetable, pycbclivefile):
    """Test reading only the loudest events from a PyCBC-Live file."""
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        loudest=True,
    )
    assert_table_equal(table, pycbclivetable.filter("snr > 500"))


def test_read_pycbc_live_extended_metadata(
    pycbclivetable,
    pycbclivepsd,
    pycbclivefile,
):
    """Test that `EventTable` can read extended metadata from a PyCBC file."""
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        extended_metadata=True,  # default
    )
    assert_table_equal(table, pycbclivetable)
    assert_array_equal(
        table.meta["loudest"],
        (pycbclivetable["snr"] > 500).nonzero()[0],
    )
    assert_quantity_sub_equal(
        table.meta["psd"],
        pycbclivepsd,
        exclude=["name", "channel", "unit", "epoch"])


def test_read_pycbc_live_extended_metadata_false(pycbclivefile):
    """Test that `EventTable` can read minimal metadata from a PyCBC file."""
    # check extended_metadata=False works
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        extended_metadata=False,
    )
    assert table.meta == {"ifo": "X1"}


def test_read_pycbc_live_multiple_ifos(
    pycbclivetable,
    pycbclivefile,
):
    """Test that `EventTable` raises an error if multiple IFOs are present."""
    with h5py.File(pycbclivefile, "r+") as h5f:
        h5f.create_group("Z1")
    with pytest.raises(
        ValueError,
        match="PyCBC live HDF5 file contains dataset groups",
    ):
        EventTable.read(pycbclivefile, format="hdf5.pycbc_live")

    # but check that we can still read the original
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        ifo="X1",
    )
    assert_table_equal(table, pycbclivetable)


def test_read_pycbc_live_processed_columns(
    pycbclivetable,
    pycbclivefile,
):
    """Test that `EventTable` can read processed columns from a PyCBC file."""
    # assert processed colums works
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        ifo="X1",
        columns=["mchirp", "new_snr"],
    )
    mchirp = (
        (pycbclivetable["mass1"] * pycbclivetable["mass2"]) ** (3/5.)
        / (pycbclivetable["mass1"] + pycbclivetable["mass2"]) ** (1/5.)
    )
    assert_array_equal(table["mchirp"], mchirp)


def test_read_pycbc_live_filter_columns(
    pycbclivetable,
    pycbclivefile,
):
    """Test that `EventTable` can read with 'where' filter and columns."""
    # test with 'where' filter and columns
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        ifo="X1",
        where="snr>.5",
        columns=("a", "b", "mass1"),
    )
    assert_table_equal(
        table,
        filter_table(pycbclivetable, "snr>.5")[("a", "b", "mass1")],
    )


def test_read_pycbc_live_regression_1081(
    pycbclivetable,
    pycbclivefile,
):
    """Test against regression of gwpy/gwpy#1081."""
    table = EventTable.read(
        pycbclivefile,
        format="hdf5.pycbc_live",
        ifo="X1",
        where="snr>.5",
        columns=("a", "b", "snr"),
    )
    assert_table_equal(
        table,
        filter_table(pycbclivetable, "snr>.5")[("a", "b", "snr")],
    )



# Copyright (c) 2026 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Test reading and writing HDF5 files using h5py."""

import h5py

from ...testing.utils import (
    TEST_HDF5_FILE,
)
from .. import (
    TimeSeries,
    TimeSeriesDict,
)

__author__ = "Duncan Macleod <macleoddm@cardiff.ac.uk>"

TEST_DATA_SPAN = (968654552.0, 968654553.0)


def test_hdf5_read():
    """Test reading HDF5 file using TimeSeries.read()."""
    data = TimeSeries.read(TEST_HDF5_FILE, path="H1:LDAS-STRAIN")
    assert data.size == 16384
    assert data.span == TEST_DATA_SPAN


def test_hdf5_read_h5file():
    """Test reading HDF5 file opened with h5py.File()."""
    with h5py.File(TEST_HDF5_FILE, "r") as h5file:
        data = TimeSeries.read(h5file, path="H1:LDAS-STRAIN")
    assert data.size == 16384
    assert data.span == TEST_DATA_SPAN


def test_hdf5_read_h5dataset():
    """Test reading HDF5 dataset using TimeSeries.read()."""
    with h5py.File(TEST_HDF5_FILE, "r") as h5file:
        dset = h5file["H1:LDAS-STRAIN"]
        data = TimeSeries.read(dset)
    assert data.size == 16384
    assert data.span == TEST_DATA_SPAN


def test_hdf5_read_values():
    """Test reading HDF5 dataset values using TimeSeries.read().

    Checks for regression against https://gitlab.com/gwpy/gwpy/-/issues/1843.
    """
    with h5py.File(TEST_HDF5_FILE, "r") as h5file:
        for ds in h5file.values():
            data = TimeSeries.read(ds)
            assert data.size == 16384
            assert data.span == TEST_DATA_SPAN


def test_hdf5_read_parallel_keys():
    """Test reading HDF5 datasets in parallel with a KeysView.

    Checks for regression against https://gitlab.com/gwpy/gwpy/-/issues/1843.
    """
    with h5py.File(TEST_HDF5_FILE, "r") as h5file:
        parallel = len(h5file)
        data = TimeSeriesDict.read(h5file, h5file.keys(), parallel=parallel)
        for key in h5file:
            ts = data[key]
            assert ts.size == 16384
            assert ts.span == TEST_DATA_SPAN



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Write GWpy objects to HDF5 files."""

from __future__ import annotations

from functools import wraps
from pathlib import Path
from typing import TYPE_CHECKING

import h5py
from astropy.io.misc.hdf5 import is_hdf5

from .utils import FileSystemPath

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import (
        Concatenate,
        Literal,
        ParamSpec,
        TypeVar,
    )

    from .utils import (
        FileLike,
        Readable,
        Writable,
    )

    P = ParamSpec("P")
    R = TypeVar("R")

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "create_dataset",
    "find_dataset",
    "identify_hdf5",
    "open_hdf5",
    "with_read_hdf5",
    "with_write_hdf5",
]


def identify_hdf5(
    origin: Literal["read", "write"],
    filepath: FileSystemPath | None = None,
    fileobj: FileLike | None = None,
    *args: object,
    **kwargs: object,
) -> bool:
    """Identify an HDF5 file based on its signature.

    Parameters
    ----------
    origin : `str`
        Either ``'read'`` or ``'write'``, indicating whether the
        identification is being done for reading or writing.

    filepath : `str`, `pathlib.Path`, optional
        The file path to check.

    fileobj : file-like, optional
        A file-like object to check.

    args, kwargs
        Any additional arguments are passed directly to
        `astropy.io.misc.hdf5.is_hdf5`.

    Returns
    -------
    is_hdf5 : `bool`
        `True` if the file is an HDF5 file, `False` otherwise.

    See Also
    --------
    astropy.io.misc.hdf5.is_hdf5
    """
    if filepath is not None:
        # is_hdf5 requires a string filepath
        filepath = str(filepath)
    return is_hdf5(origin, filepath, fileobj, *args, **kwargs)


def open_hdf5(
    source: FileLike | h5py.HLObject,
    mode: str = "r",
    **kwargs,
) -> h5py.HLObject:
    """Open a :class:`h5py.File` from disk, gracefully handling corner cases.

    If ``source`` is already an HDF5 object, it is simply returned.
    Otherwise, it is opened as an HDF5 file in the given mode.

    Parameters
    ----------
    source : file-like, `h5py.Group`, `h5py.Dataset`
        The file path or file-like object to open, or an existing
        HDF5 object to return as-is.

    mode : `str`
        The mode in which to open the file, default: ``'r'``.

    kwargs
        Any additional keyword arguments are passed directly to
        :class:`h5py.File`.

    Returns
    -------
    h5obj : `h5py.File`, `h5py.Group`, `h5py.Dataset`
        The opened HDF5 file object, or the original object if it was
        already an HDF5 object.
    """
    if isinstance(source, h5py.HLObject):
        return source
    return h5py.File(source, mode=mode, **kwargs)


def with_read_hdf5(
    func: Callable[Concatenate[Readable | h5py.HLObject, P], R],
) -> Callable[Concatenate[h5py.HLObject, P], R]:
    """Decorate an HDF5-reading function to open a filepath if needed.

    The decorated function will accept file paths or readable objects,
    but will always pass an h5py.Group as the first argument to the
    original function.

    Parameters
    ----------
    func : callable
        Function that expects an h5py.Group as its first argument

    Returns
    -------
    callable
        Decorated function that accepts file paths/readable objects
    """
    @wraps(func)
    def decorated_func(
        fobj: Readable,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> R:
        if isinstance(fobj, h5py.HLObject):
            return func(fobj, *args, **kwargs)
        with h5py.File(fobj, "r") as h5f:
            return func(h5f, *args, **kwargs)

    return decorated_func


def find_dataset(
    h5o: h5py.File | h5py.Group | h5py.Dataset,
    path: str | None = None,
) -> h5py.Dataset:
    """Find and return the relevant dataset inside the given H5 object.

    If ``path=None`` is given, and ``h5o`` contains a single dataset, that
    will be returned

    Parameters
    ----------
    h5o : `h5py.File`, `h5py.Group`
        the HDF5 object in which to search

    path : `str`, optional
        the path (relative to ``h5o``) of the desired data set

    Returns
    -------
    dset : `h5py.Dataset`
        the recovered dataset object

    Raises
    ------
    ValueError
        if ``path=None`` and the HDF5 object contains multiple datasets
    KeyError
        if ``path`` is given but is not found within the HDF5 object
    """
    # find dataset
    if isinstance(h5o, h5py.Dataset):
        return h5o
    if path is None and len(h5o) == 1:
        path = next(iter(h5o.keys()))
    elif path is None:
        msg = "Please specify the HDF5 path via the ``path=`` keyword argument"
        raise ValueError(msg)
    return h5o[path]


# -- writing utilities ---------------

def with_write_hdf5(func: Callable[..., object]) -> Callable[..., object]:
    """Decorate an HDF5-writing function to open a filepath if needed.

    ``func`` should be written to take the object to be written as the
    first argument, and then presume an `h5py.Group` as the second.

    This method uses keywords ``append`` and ``overwrite`` as follows if
    the output file already exists:

    - ``append=False, overwrite=False``: raise `~exceptions.IOError`
    - ``append=True``: open in mode ``a``
    - ``append=False, overwrite=True``: open in mode ``w``
    """
    @wraps(func)
    def decorated_func(
        obj: object,
        fobj: Writable | h5py.HLObject,
        *args: object,
        **kwargs: object,
    ) -> object:
        if not isinstance(fobj, h5py.HLObject):
            append = kwargs.get("append", False)
            overwrite = kwargs.get("overwrite", False)
            # Check if file exists (only for path-like objects)
            if (
                isinstance(fobj, FileSystemPath)
                and Path(fobj).exists()
                and not (overwrite or append)
            ):
                msg = f"File exists: {fobj}"
                raise OSError(msg)
            with h5py.File(fobj, "a" if append else "w") as h5f:
                return func(obj, h5f, *args, **kwargs)
        return func(obj, fobj, *args, **kwargs)

    return decorated_func


def create_dataset(
    parent: h5py.Group | h5py.File,
    path: str,
    *,
    overwrite: bool = False,
    **kwargs: object,
) -> h5py.Dataset:
    """Create a new dataset inside the parent HDF5 object.

    Parameters
    ----------
    parent : `h5py.Group`, `h5py.File`
        the object in which to create a new dataset

    path : `str`
        the path at which to create the new dataset

    overwrite : `bool`
        if `True`, delete any existing dataset at the desired path,
        default: `False`

    **kwargs
        other arguments are passed directly to
        :meth:`h5py.Group.create_dataset`

    Returns
    -------
    dataset : `h5py.Dataset`
        the newly created dataset
    """
    # force deletion of existing dataset
    if path in parent and overwrite:
        del parent[path]

    # create new dataset with improved error handling
    try:
        return parent.create_dataset(path, **kwargs)
    except RuntimeError as exc:
        if str(exc) == "Unable to create link (Name already exists)":
            msg = (
                f"{exc}: '{path}', pass overwrite=True to ignore "
                f"existing datasets"
            )
            exc.args = (msg,)
        raise



# Copyright (c) 2021-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Custom pytest marks for GWpy.

This is mainly to allow downstream users/builders/packagers to skip tests
that require annoying/flaky configuration, or just take too long.

This module is imported in :mod:`gwpy.conftest` such that all marks are
registered up front and visible via ``python -m pytest gwpy --markers``.
"""

MARKS: dict[str, str] = {
    # nothing here, used to support 'cvmfs' but that was removed
}


def _register_marks(config):
    """Register all marks for GWpy.

    This function is designed to be called from :mod:`gwpy.conftest`.
    """
    for name, doc in MARKS.items():
        config.addinivalue_line(
            "markers",
            f"{name}: {doc}",
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""The `Array`.

The `Array` structure provides the core array-with-metadata environment
with the standard array methods wrapped to return instances of itself.

Each sub-class of `Array` should override the `Array._metadata_slots`
attribute, giving a list of the valid properties for these data. This is
critical to being able to view data with this class, used when copying and
transforming instances of the class.
"""

from __future__ import annotations

import contextlib
import copy
import os
import textwrap
from decimal import Decimal
from math import modf
from typing import TYPE_CHECKING

import numpy
from astropy.units import Quantity
from astropy.utils.compat.numpycompat import COPY_IF_NEEDED

from ..detector import Channel
from ..detector.units import parse_unit
from ..time import Time, to_gps

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Iterable,
    )
    from typing import (
        ClassVar,
        Literal,
        Self,
    )

    from astropy.units import UnitBase
    from astropy.units.typing import QuantityLike
    from numpy.typing import DTypeLike

    from ..time import SupportsToGps
    from ..typing import UnitLike
    from .sliceutils import SliceLike

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

numpy.set_printoptions(threshold=200, linewidth=65)


# -- core Array ----------------------

class Array(Quantity):
    """Array holding data with a unit, and other metadata.

    This `Array` holds the input data and a standard set of metadata
    properties associated with GW data.

    Parameters
    ----------
    value : array-like
        Input data array.

    unit : `~astropy.units.Unit`
        Physical unit of these data.

    epoch : `~gwpy.time.LIGOTimeGPS`, `float`, `str`
        GPS epoch associated with these data,
        any input parsable by `~gwpy.time.to_gps` is fine

    name : `str`
        Descriptive title for this array.

    channel : `~gwpy.detector.Channel`, `str`
        Source data stream for these data.

    dtype : `~numpy.dtype`
        Input data type.

    copy : `bool`
        Choose to copy the input data to new memory.

    subok : `bool`
        Allow passing of sub-classes by the array generator.

    Returns
    -------
    array : `Array`
        A new array, with a view of the data, and all associated metadata.

    Examples
    --------
    To create a new `Array` from a list of samples:

    >>> a = Array([1, 2, 3, 4, 5], 'm/s', name='my data')
    >>> print(a)
    Array([ 1., 2., 3., 4., 5.]
          unit: Unit("m / s"),
          name: 'my data',
          epoch: None,
          channel: None)
    """

    #: list of new attributes defined in this class
    #
    # this is used in __array_finalize__ to create new instances of this
    # object [http://docs.scipy.org/doc/numpy/user/basics.subclassing.html]
    _metadata_slots: ClassVar[tuple[str, ...]] = (
        "name",
        "epoch",
        "channel",
    )

    def __new__(
        cls,
        value: QuantityLike,
        *,
        # Quantity attrs
        unit: UnitBase | str | None = None,
        # new attrs
        name: str | None = None,
        epoch: SupportsToGps | None = None,
        channel: Channel | str | None = None,
        # ndarray attrs
        dtype: DTypeLike = None,
        copy: bool = True,
        subok: bool = True,
        order: str | None = None,
        ndmin: int = 0,
    ) -> Self:
        """Create a new `Array`."""
        # pick dtype from input array
        if dtype is None and isinstance(value, numpy.ndarray):
            dtype = value.dtype

        # parse unit with forgiveness
        if unit is not None:
            unit = parse_unit(unit, parse_strict="warn")

        # create new array
        new = super().__new__(
            cls,
            value,
            unit=unit,
            dtype=dtype,
            copy=COPY_IF_NEEDED,
            order=order,
            subok=subok,
            ndmin=ndmin,
        )

        # explicitly copy here to get ownership of the data,
        # see (astropy/astropy#7244)
        if copy:
            new = new.copy()

        # set new attributes
        if name is not None:
            new.name = name
        if epoch is not None:
            new.epoch = epoch
        if channel is not None:
            new.channel = channel

        return new

    # -- object creation -------------
    # methods here handle how these objects are created,
    # mainly to do with making sure metadata attributes get
    # properly reassigned from old to new

    def _wrap_function(
        self,
        function: Callable,
        *args,  # noqa: ANN002
        **kwargs,
    ) -> Self | Quantity:
        # if the output of the function is a scalar, return it as a Quantity
        # not whatever class this is
        out = super()._wrap_function(function, *args, **kwargs)
        if out.ndim == 0:
            return Quantity(out.value, out.unit)
        return out

    def __quantity_subclass__(
        self,
        unit: UnitBase,
    ) -> tuple[type, Literal[True]]:
        """View operations should return the same type (or a subclass)."""
        return type(self), True

    def __array_finalize__(self, obj: Self | None) -> None:
        """Finalise this `Array`.

        This is called whenever a new view of an `Array` is created.
        """
        # format a new instance of this class starting from `obj`
        if obj is None:
            return

        # call Quantity.__array_finalize__ to handle the units
        super().__array_finalize__(obj)

        # then update metadata
        if isinstance(obj, Quantity):
            self.__metadata_finalize__(obj, force=False)

    def __metadata_finalize__(
        self,
        obj: Self,
        *,
        force: bool = False,
    ) -> None:
        """Finalise metadata for this `Array`."""
        # apply metadata from obj to self if creating a new object
        for attr in self._metadata_slots:
            _attr = f"_{attr}"  # use private attribute (not property)
            # if attribute is unset, default it to None, then update
            # from obj if desired
            try:
                getattr(self, _attr)
            except AttributeError:
                update = True
            else:
                update = force
            if update:
                try:
                    val = getattr(obj, _attr)
                except AttributeError:
                    continue
                else:
                    if isinstance(val, Quantity):  # copy Quantities
                        setattr(self, _attr, type(val)(val))
                    else:
                        setattr(self, _attr, val)

    def __getitem__(
        self,
        item: SliceLike | tuple[SliceLike, ...],
    ) -> Self | Quantity:
        """Get an item from this `Array`.

        This implements ``array[item]``.
        """
        new = super().__getitem__(item)

        # return scalar as a Quantity
        if numpy.ndim(new) == 0:
            return Quantity(new, unit=self.unit)

        return new

    # -- display ---------------------

    def _repr_helper(self, print_: Callable) -> str:
        """Create a string representation of an `Array`."""
        if print_ is repr:
            opstr = "="
        else:
            opstr = ": "

        # get prefix and suffix
        prefix = f"{type(self).__name__}("
        suffix = ")"
        if print_ is repr:
            prefix = f"<{prefix}"
            suffix += ">"

        indent = " " * len(prefix)

        # format value
        arrstr = numpy.array2string(
            self.value,
            separator=", ",
            prefix=prefix,
        )

        # format unit
        metadata = [("unit", print_(self.unit) or "dimensionless")]

        # format other metadata
        try:
            attrs = self._print_slots
        except AttributeError:
            attrs = self._metadata_slots
        for attr in attrs:
            try:
                val = getattr(self, attr)
            except (AttributeError, KeyError):
                val = None
            thisindent = indent + " " * (len(attr) + len(opstr))
            metadata.append((
                attr.lstrip("_"),
                textwrap.indent(print_(val), thisindent).strip(),
            ))
        metadatastr = f",{os.linesep}{indent}".join(
            f"{attr}{opstr}{value}" for attr, value in metadata
        )

        return f"{prefix}{arrstr},{os.linesep}{indent}{metadatastr}{suffix}"

    def __repr__(self) -> str:
        """Return a representation of this object.

        This just represents each of the metadata objects appropriately
        after the core data array
        """
        return self._repr_helper(repr)

    def __str__(self) -> str:
        """Return a printable string format representation of this object.

        This just prints each of the metadata objects appropriately
        after the core data array
        """
        return self._repr_helper(str)

    # -- new properties --------------

    # name
    @property
    def name(self) -> str | None:
        """Name for this data set."""
        self._name: str | None
        try:
            return self._name
        except AttributeError:
            self._name = None
            return self._name

    @name.setter
    def name(self, val: str | None) -> None:
        if val is None:
            self._name = None
        else:
            self._name = str(val)

    @name.deleter
    def name(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._name

    # epoch
    @property
    def epoch(self) -> Time | None:
        """GPS epoch associated with these data."""
        self._epoch: Time | None
        try:
            if self._epoch is None:
                return None
            return Time(*modf(self._epoch)[::-1], format="gps", scale="utc")
        except AttributeError:
            self._epoch = None
            return self._epoch

    @epoch.setter
    def epoch(self, epoch: SupportsToGps | None) -> None:
        if epoch is None:
            self._epoch = None
        else:
            self._epoch = Decimal(str(to_gps(epoch)))

    @epoch.deleter
    def epoch(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._epoch

    # channel
    @property
    def channel(self) -> Channel | None:
        """Instrumental channel associated with these data."""
        self._channel: Channel | None
        try:
            return self._channel
        except AttributeError:
            self._channel = None
            return self._channel

    @channel.setter
    def channel(self, chan: Channel | str | None) -> None:
        if isinstance(chan, Channel):
            self._channel = chan
        elif chan is None:
            self._channel = None
        else:
            self._channel = Channel(chan)

    @channel.deleter
    def channel(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._channel

    # unit - we override this to make the property less pedantic
    #        astropy won't allow you to set a unit that it doesn't
    #        recognise
    @property
    def unit(self) -> UnitBase | None:
        """The physical unit of these data."""
        try:
            return self._unit
        except AttributeError:
            return None

    @unit.setter
    def unit(self, unit: UnitBase | str | None) -> None:
        if getattr(self, "_unit", None) is not None:
            msg = (
                "can't set attribute; to change the units of this "
                f"{type(self).__name__}, use the .to() instance method "
                "instead, otherwise use the override_unit() instance method "
                "to forcefully set a new unit",
            )
            raise AttributeError(msg)
        self._unit = parse_unit(unit)

    @unit.deleter
    def unit(self) -> None:
        with contextlib.suppress(AttributeError):
            del self._unit

    # -- array methods ---------------

    def __array_ufunc__(
        self,
        function: Callable,
        method: str,
        *inputs,  # noqa: ANN002
        **kwargs,
    ) -> Self | Quantity:
        """Wrap a ufunc, handling units, and scalar outputs."""
        out = super().__array_ufunc__(function, method, *inputs, **kwargs)
        # if a ufunc returns a scalar, return a Quantity
        if not out.ndim:
            return Quantity(out, copy=COPY_IF_NEEDED)
        # otherwise return an array
        return out

    def abs(
        self,
        **kwargs,
    ) -> Self | Quantity:
        """Return the absolute value of the data in this `Array`.

        See Also
        --------
        numpy.absolute
            For details of all available positional and keyword arguments,
            and for details of the return value.
        """
        return self._wrap_function(
            numpy.absolute,
            **kwargs,
        )

    def median(
        self,
        axis: int | Iterable[int] | None = None,
        **kwargs,
    ) -> Self | Quantity:
        """Return the median of the data in this `Array`.

        See Also
        --------
        numpy.median
            For details of all available positional and keyword arguments,
            and for details of the return value.
        """
        return self._wrap_function(
            numpy.median,
            axis=axis,
            **kwargs,
        )

    def _to_own_unit(
        self,
        value: QuantityLike,
        *,
        check_precision: bool = True,
        unit: UnitBase | None = None,
    ) -> Self:
        if unit is None and self.unit is None:
            unit = ""
        return super()._to_own_unit(
            value,
            check_precision=check_precision,
            unit=unit,
        )

    _to_own_unit.__doc__ = Quantity._to_own_unit.__doc__  # noqa: SLF001

    def override_unit(
        self,
        unit: UnitLike,
        parse_strict: Literal["raise", "warn", "silent"] = "raise",
    ) -> None:
        """Reset the unit of these data.

        Use of this method is discouraged in favour of `to()`,
        which performs accurate conversions from one unit to another.
        The method should really only be used when the original unit of the
        array is plain wrong.

        Parameters
        ----------
        unit : `~astropy.units.Unit`, `str`
            the unit to force onto this array

        parse_strict : `str`
            how to handle errors in the unit parsing, default is to
            raise the underlying exception from `astropy.units`

        See Also
        --------
        gwpy.detector.units.parse_unit
            For details of unit string parsing.
        """
        self._unit = parse_unit(unit, parse_strict=parse_strict)

    def flatten(self, order: str = "C") -> Quantity:
        """Return a copy of the array collapsed into one dimension.

        Any index information is removed as part of the flattening,
        and the result is returned as a `~astropy.units.Quantity` array.

        Parameters
        ----------
        order : {'C', 'F', 'A', 'K'}
            'C' means to flatten in row-major (C-style) order.
            'F' means to flatten in column-major (Fortran-
            style) order. 'A' means to flatten in column-major
            order if `a` is Fortran *contiguous* in memory,
            row-major order otherwise. 'K' means to flatten
            `a` in the order the elements occur in memory.
            The default is 'C'.

        Returns
        -------
        y : `~astropy.units.Quantity`
            A copy of the input array, flattened to one dimension.

        See Also
        --------
        ravel : Return a flattened array.
        flat : A 1-D flat iterator over the array.

        Examples
        --------
        >>> a = Array([[1,2], [3,4]], unit='m', name='Test')
        >>> a.flatten()
        <Quantity [1., 2., 3., 4.] m>
        """
        return super().flatten(order=order).view(Quantity)

    def copy(self, order: Literal["C", "F", "A", "K"] = "C") -> Self:
        """Return a copy of this `Array`.

        Parameters
        ----------
        order : {'C', 'F', 'A', 'K'}, optional
            The desired memory layout order for the copy.

        See Also
        --------
        numpy.ndarray.copy
            For details of the copy operation.
        """
        out = super().copy(order=order)
        for slot in self._metadata_slots:
            old = getattr(self, f"_{slot}", None)
            if old is not None:
                setattr(out, slot, copy.copy(old))
        return out

    copy.__doc__ = Quantity.copy.__doc__

    # -- type helpers ----------------

    def __mul__(
        self,
        other: QuantityLike,
    ) -> Self:
        """Multiply this `Array` by another quantity."""
        return super().__mul__(other)

    def __pow__(
        self,
        power: QuantityLike,
    ) -> Self:
        """Raise this `Array` to a power."""
        return super().__pow__(power)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Basic utilities for reading/writing LIGO_LW-format XML files.

All specific unified input/output for class objects should be placed in
an 'io' subdirectory of the containing directory for that class.
"""


from __future__ import annotations

import os
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    cast,
)

import numpy

from .utils import (
    FileLike,
    file_list,
)

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Collection,
        Generator,
        Iterable,
    )
    from typing import Literal
    from xml.sax.handler import ContentHandler
    from xml.sax.xmlreader import AttributesImpl

    from igwn_ligolw.ligolw import (
        Document,
        Element,
        FilteringLIGOLWContentHandler,
        PartialLIGOLWContentHandler,
        Stream,
        Table,
    )

    from .utils import (
        FileSystemPath,
        NamedReadable,
        Readable,
        Writable,
    )

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

__all__ = [
    "build_content_handler",
    "default_content_handler",
    "get_filtering_contenthandler",
    "get_ligolw_element",
    "get_partial_contenthandler",
    "is_ligolw",
    "iter_tables",
    "list_tables",
    "open_xmldoc",
    "read_ligolw",
    "read_table",
    "to_table_type",
    "write_tables",
    "write_tables_to_document",
]

# XML elements
XML_SIGNATURE = b"<?xml"
LIGOLW_SIGNATURE = b"<!doctype ligo_lw"
LIGOLW_ELEMENT = b"<ligo_lw>"


# -- content handling ----------------

def _int_ilwd(ilwd: str) -> int:
    """Convert an ``ilwd`` string into an integer."""
    try:
        _, _, i = ilwd.strip().split(":")
    except ValueError as exc:
        msg = f"invalid ilwd:char '{ilwd}'"
        raise ValueError(msg) from exc
    return int(i)


def strip_ilwdchar(content_handler: type[ContentHandler]) -> type[ContentHandler]:
    """Wrap a contenthandler to swap ilwdchar for int when reading a document.

    This is adapted from :func:`ligo.skymap.utils.ilwd`, copyright
    Leo Singer (GPL-3.0-or-later).
    """
    from igwn_ligolw.ligolw import (
        Column,
        Table,
    )
    from igwn_ligolw.lsctables import TableByName
    from igwn_ligolw.types import FromPyType

    class IlwdMapContentHandler(content_handler):  # type: ignore[misc,valid-type]

        def __init__(self, *args, **kwargs) -> None:  # noqa: ANN002
            super().__init__(*args, **kwargs)
            self._idconverter: dict[tuple[int, str], Callable[[str], int]] = {}

        def startColumn(  # noqa: N802
            self,
            parent: Element,
            attrs: AttributesImpl,
        ) -> Column:
            result = super().startColumn(parent, attrs)

            # if an old ID type, convert type definition to an int
            if result.Type == "ilwd:char":
                self._idconverter[(id(parent), result.Name)] = _int_ilwd
                result.Type = FromPyType[int]

            try:
                validcolumns = TableByName[parent.Name].validcolumns
            except KeyError:  # parent.Name not in TableByName
                return result
            if result.Name not in validcolumns:
                stripped_column_to_valid_column = {
                    Column.ColumnName(name): name
                    for name in validcolumns
                }
                if result.Name in stripped_column_to_valid_column:
                    result.setAttribute(
                        "Name",
                        stripped_column_to_valid_column[result.Name],
                    )

            return result

        def startStream(  # noqa: N802
            self,
            parent: Element,
            attrs: AttributesImpl,
        ) -> Stream:
            result = super().startStream(parent, attrs)
            if isinstance(result, Table.Stream):
                loadcolumns = set(parent.columnnames)
                if parent.loadcolumns is not None:
                    loadcolumns &= set(parent.loadcolumns)
                pid = id(parent)
                result._tokenizer.set_types([
                    self._idconverter.pop((pid, colname), pytype)
                    if colname in loadcolumns else None
                    for pytype, colname in zip(
                        parent.columnpytypes,
                        parent.columnnames,
                        strict=True,
                    )
                ])
            return result

    return IlwdMapContentHandler


def _wrap_content_handler(contenthandler: type[ContentHandler]) -> type[ContentHandler]:

    @strip_ilwdchar
    class ContentHandler(contenthandler):  # type: ignore[misc,valid-type]
        pass

    return ContentHandler


def default_content_handler() -> type[ContentHandler]:
    """Return a standard content handler to read LIGO_LW documents.

    This handler knows how to parse LSCTables, and automatically converts
    old-style ilwdchar ID types to `int`.

    Returns
    -------
    contenthandler : subclass of `igwn_ligolw.ligolw.LIGOLWContentHandler`
    """
    from igwn_ligolw.ligolw import LIGOLWContentHandler
    return _wrap_content_handler(LIGOLWContentHandler)


def get_partial_contenthandler(
    element: type[Element],
) -> type[PartialLIGOLWContentHandler]:
    """Build a `PartialLIGOLWContentHandler` to read only this element.

    Parameters
    ----------
    element : `type`, subclass of :class:`~igwn_ligolw.ligolw.Element`
        the element class to be read

    Returns
    -------
    contenthandler : `type`
        a subclass of `~igwn_ligolw.ligolw.PartialLIGOLWContentHandler`
        to read only the given `element`
    """
    from igwn_ligolw.ligolw import (
        PartialLIGOLWContentHandler,
        Table,
    )

    if issubclass(element, Table):
        def _element_filter(name: str, attrs: AttributesImpl) -> bool:
            return element.CheckProperties(name, attrs)
    else:
        def _element_filter(name: str, attrs: AttributesImpl) -> bool:  # noqa: ARG001
            return name == element.tagName

    return build_content_handler(PartialLIGOLWContentHandler, _element_filter)


def get_filtering_contenthandler(
    element: type[Element],
) -> type[FilteringLIGOLWContentHandler]:
    """Build a `FilteringLIGOLWContentHandler` to exclude this element.

    Parameters
    ----------
    element : `type`, subclass of :class:`~igwn_ligolw.ligolw.Element`
        the element to exclude (and its children)

    Returns
    -------
    contenthandler : `type`
        a subclass of `~igwn_ligolw.ligolw.FilteringLIGOLWContentHandler`
        to exclude an element and its children
    """
    from igwn_ligolw.ligolw import (
        FilteringLIGOLWContentHandler,
        Table,
    )

    if issubclass(element, Table):
        def _element_filter(name: str, attrs: AttributesImpl) -> bool:
            return ~element.CheckProperties(name, attrs)
    else:
        def _element_filter(name: str, attrs: AttributesImpl) -> bool:  # noqa: ARG001
            return name != element.tagName

    return build_content_handler(
        FilteringLIGOLWContentHandler,
        _element_filter,
    )


def build_content_handler(
    parent: type[PartialLIGOLWContentHandler],
    filter_func: Callable[[str, AttributesImpl], bool],
) -> type[PartialLIGOLWContentHandler]:
    """Build a `~xml.sax.handler.ContentHandler` with a given filter.

    Parameters
    ----------
    parent : `type`, subclass of `xml.sax.handler.ContentHandler`
        a class of contenthandler to use

    filter_func : `callable`
        the filter function to pass to the content handler creation

    Returns
    -------
    contenthandler : subclass of ``parent``
        a new content handler that applies the filter function and the
        default parsing extras from :func:`_wrap_content_handler`.
    """
    class ContentHandler(parent):  # type: ignore[misc,valid-type]
        def __init__(self, document: Document) -> None:
            super().__init__(document, filter_func)

    return _wrap_content_handler(ContentHandler)


# -- reading -------------------------

def read_ligolw(
    source: NamedReadable | list[NamedReadable],
    contenthandler: type[ContentHandler] | None = None,
    **kwargs,
) -> Document:
    """Read one or more LIGO_LW format files.

    Parameters
    ----------
    source : `str`, `file`, `list` of `str` or `file`
        The open file or file path to read.

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        Content handler used to parse document.

    kwargs
        Other keyword arguments to pass to `igwn_ligolw.utils.load_url`.

    Returns
    -------
    xmldoc : `~igwn_ligolw.ligolw.Document`
        the document object as parsed from the file(s)
    """
    from igwn_ligolw import types
    from igwn_ligolw.ligolw import Document
    from igwn_ligolw.utils import ligolw_add, load_url

    # mock ToPyType to link to numpy dtypes
    topytype = types.ToPyType.copy()
    for key in types.ToPyType:
        if key in types.ToNumPyType:
            types.ToPyType[key] = numpy.dtype(types.ToNumPyType[key]).type

    # set contenthandler
    if contenthandler is None:
        contenthandler = default_content_handler()

    # read one or more files into a single Document
    sources = file_list(source)
    try:
        if len(sources) == 1:
            return load_url(
                sources[0],
                contenthandler=contenthandler,
                **kwargs,
            )
        return ligolw_add.ligolw_add(
            Document(),
            sources,
            contenthandler=contenthandler,
            **kwargs,
        )
    finally:  # replace ToPyType
        types.ToPyType = topytype


# -- reading -------------------------

def read_table(
    source: Document | NamedReadable | list[NamedReadable],
    tablename: str | None = None,
    columns: Collection[str] | None = None,
    contenthandler: type[ContentHandler] | None = None,
    **kwargs,
) -> Table:
    """Read a :class:`~igwn_ligolw.ligolw.Table` from one or more LIGO_LW files.

    Parameters
    ----------
    source : `Document`, `file`, `str`, `CacheEntry`, `list`
        object representing one or more files. One of

        - a LIGO_LW :class:`~igwn_ligolw.ligolw.Document`
        - an open `file`
        - a `str` pointing to a file path on disk
        - a formatted :class:`~lal.utils.CacheEntry` representing one file
        - a `list` of `str` file paths or :class:`~lal.utils.CacheEntry`

    tablename : `str`
        name of the table to read.

    columns : `list`, optional
        list of column name strings to read, default all.

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        SAX content handler for parsing LIGO_LW documents.

    **kwargs
        other keyword arguments are passed to `~gwpy.io.ligolw.read_ligolw`

    Returns
    -------
    table : :class:`~igwn_ligolw.ligolw.Table`
        `Table` of data
    """
    from igwn_ligolw import lsctables
    from igwn_ligolw.ligolw import (
        Document,
        Table,
    )

    # get content handler to read only this table (if given)
    if tablename is not None:
        tableclass = lsctables.TableByName[
            Table.TableName(tablename)
        ]
        if contenthandler is None:
            contenthandler = get_partial_contenthandler(tableclass)

        # overwrite loading column names to get just what was asked for
        _oldcols = tableclass.loadcolumns
        if columns is not None:
            tableclass.loadcolumns = set(columns)

    # read document
    if isinstance(source, Document):
        xmldoc = source
    else:
        if contenthandler is None:
            contenthandler = default_content_handler()
        try:
            xmldoc = read_ligolw(
                source,
                contenthandler=contenthandler,
                **kwargs)
        finally:  # reinstate original set of loading column names
            if tablename is not None:
                tableclass.loadcolumns = _oldcols

    # now find the right table
    if tablename is None:
        tables = list_tables(xmldoc)
        if not tables:
            msg = "No tables found in LIGO_LW document(s)"
            raise ValueError(msg)
        if len(tables) > 1:
            msg = (
                "Multiple tables found in LIGO_LW document(s), please specify "
                "the table to read via the ``tablename=`` keyword argument. "
                "The following tables were found: "
                "'{}'".format("', '".join(tables))
            )
            raise ValueError(msg)
        tableclass = lsctables.TableByName[Table.TableName(tables[0])]

    # extract table
    return tableclass.get_table(xmldoc)


# -- writing -------------------------

def open_xmldoc(
    fobj: Readable,
    contenthandler: type[ContentHandler] | None = None,
    **kwargs,
) -> Document:
    """Try and open an existing LIGO_LW-format file, or create a new Document.

    Parameters
    ----------
    fobj : `str`, `file`
        file path or open file object to read

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        the content handler with which to parse the document, if not given
        a default handler will be created using
        :func:`default_content_handler`.

    **kwargs
        other keyword arguments to pass to
        :func:`~igwn_ligolw.utils.load_fileobj` as appropriate

    Returns
    -------
    xmldoc : :class:`~igwn_ligolw.ligolw.Document`
        either the `Document` as parsed from an existing file, or a new, empty
        `Document`
    """
    from igwn_ligolw.ligolw import Document
    from igwn_ligolw.utils import load_fileobj

    if contenthandler is None:
        contenthandler = default_content_handler()

    # read from an existing Path/filename
    if not isinstance(fobj, FileLike):
        try:
            with open(fobj, "rb") as fobj2:
                return open_xmldoc(
                    fobj2,
                    contenthandler=contenthandler,
                    **kwargs,
                )
        except OSError:
            # or just create a new Document
            return Document()

    return load_fileobj(
        fobj,
        contenthandler=contenthandler,
        **kwargs,
    )


def get_ligolw_element(xmldoc: Document) -> Element:
    """Find an existing <LIGO_LW> element in this XML Document."""
    from igwn_ligolw.ligolw import LIGO_LW, WalkChildren

    if isinstance(xmldoc, LIGO_LW):
        return xmldoc
    for elem in WalkChildren(xmldoc):
        if isinstance(elem, LIGO_LW):
            return elem
    msg = "Cannot find LIGO_LW element in XML Document"
    raise ValueError(msg)


def write_tables_to_document(
    xmldoc: Document,
    tables: Iterable[Table],
    *,
    overwrite: bool = False,
) -> Document:
    """Write the given LIGO_LW table into a :class:`Document`.

    Parameters
    ----------
    xmldoc : :class:`~igwn_ligolw.ligolw.Document`
        the document to write into

    tables : `list` of :class:`~igwn_ligolw.ligolw.Table`
        the set of tables to write

    overwrite : `bool`, optional, default: `False`
        if `True`, delete an existing instance of the table type, otherwise
        append new rows
    """
    from igwn_ligolw import lsctables
    from igwn_ligolw.ligolw import LIGO_LW

    # find or create LIGO_LW tag
    try:
        llw = get_ligolw_element(xmldoc)
    except ValueError:
        llw = LIGO_LW()
        xmldoc.appendChild(llw)

    for table in tables:
        try:  # append new data to existing table
            old = lsctables.TableByName[
                table.TableName(table.Name)].get_table(xmldoc)
        except ValueError:
            # or create a new table
            llw.appendChild(table)
        else:
            if overwrite:
                llw.removeChild(old)
                old.unlink()
                llw.appendChild(table)
            else:
                old.extend(table)

    return xmldoc


def write_tables(
    target: Writable | Document,
    tables: Iterable[Table],
    *,
    append: bool = False,
    overwrite: bool = False,
    contenthandler: type[ContentHandler] | None = None,
    **kwargs,
) -> None:
    """Write an LIGO_LW table to file.

    Parameters
    ----------
    target : `str`, `file`, :class:`~igwn_ligolw.ligolw.Document`
        the file or document to write into

    tables : `list`, `tuple` of :class:`~igwn_ligolw.ligolw.Table`
        the tables to write

    append : `bool`, optional, default: `False`
        if `True`, append to an existing file/table, otherwise `overwrite`

    overwrite : `bool`, optional, default: `False`
        if `True`, delete an existing instance of the table type, otherwise
        append new rows

    contenthandler : `~xml.sax.handler.ContentHandler`, optional
        the content handler with which to parse the document, if not given
        a default handler will be created using
        :func:`default_content_handler`.

    **kwargs
        other keyword arguments to pass to
        :func:`~igwn_ligolw.utils.load_fileobj` as appropriate
    """
    from igwn_ligolw import utils as ligolw_utils
    from igwn_ligolw.ligolw import LIGO_LW, Document

    # allow writing directly to XML
    if isinstance(target, Document | LIGO_LW):
        xmldoc = target
    # open existing document, if possible
    elif append:
        if contenthandler is None:
            contenthandler = default_content_handler()
        xmldoc = open_xmldoc(
            target,
            contenthandler=contenthandler,
        )
    # fail on existing document and not overwriting
    elif (
        not overwrite
        and isinstance(target, str | os.PathLike)
        and Path(target).exists()
    ):
        msg = f"File exists: {target}"
        raise OSError(msg)
    else:  # or create a new document
        xmldoc = Document()

    # convert table to format
    write_tables_to_document(xmldoc, tables, overwrite=overwrite)

    # find writer function and target filename
    if isinstance(target, FileLike):
        writer = ligolw_utils.write_fileobj
        try:
            name = target.name  # type: ignore[union-attr]
        except AttributeError:
            name = ""
    else:
        writer = ligolw_utils.write_filename
        name = target = str(target)

    # handle gzip compression kwargs
    if name.endswith(".gz"):
        kwargs.setdefault("compress", "gz")

    # write XML
    writer(xmldoc, target, **kwargs)


# -- utilities -----------------------

def iter_tables(
    source: Document | NamedReadable | list[NamedReadable],
) -> Generator[Table, None, None]:
    """Iterate over all tables in the given document(s).

    Parameters
    ----------
    source : `file`, `str`, :class:`~igwn_ligolw.ligolw.Document`, `list`
        One or more open files, file paths,
        or `LIGO_LW documents <igwn_ligolw.ligolw.Document>`.

    Yields
    ------
    igwn_ligolw.ligolw.Table
        A table structure from the document(s).
    """
    from igwn_ligolw.ligolw import Element, Stream, WalkChildren

    # get LIGO_LW object
    if not isinstance(source, Element):
        filt = get_filtering_contenthandler(Stream)
        source = read_ligolw(source, contenthandler=filt)
    llw = get_ligolw_element(source)

    # yield tables
    for elem in WalkChildren(llw):
        if elem.tagName == "Table":
            yield elem


def list_tables(source: FileLike | str | Document | list) -> list[str]:
    """List the names of all tables in this file(s).

    Parameters
    ----------
    source : `file`, `str`, :class:`~igwn_ligolw.ligolw.Document`, `list`
        One or more open files, file paths,
        or `LIGO_LW documents <igwn_ligolw.ligolw.Document>`.

    Examples
    --------
    >>> from gwpy.io.ligolw import list_tables
    >>> print(list_tables('H1-LDAS_STRAIN-968654552-10.xml.gz'))
    ['process', 'process_params', 'sngl_burst', 'search_summary', 'segment_definer', 'segment_summary', 'segment']
    """  # noqa: E501
    return [tbl.TableName(tbl.Name) for tbl in iter_tables(source)]


def to_table_type(
    val: object,
    cls: type[Table],
    colname: str,
) -> object:
    """Cast a value to the correct type for inclusion in a LIGO_LW table.

    This method returns the input unmodified if a type mapping for ``colname``
    isn't found.

    Parameters
    ----------
    val : `object`
        The input object to convert, of any type

    cls : `type`, subclass of :class:`~igwn_ligolw.ligolw.Table`
        the table class to map against

    colname : `str`
        The name of the mapping column

    Returns
    -------
    obj : `object`
        The input ``val`` cast to the correct type

    Examples
    --------
    >>> from gwpy.io.ligolw import to_table_type as to_ligolw_type
    >>> from igwn_ligolw.lsctables import SnglBurstTable
    >>> x = to_ligolw_type(1.0, SnglBurstTable, 'central_freq'))
    >>> print(type(x), x)
    <class 'numpy.float32'> 1.0
    """
    from igwn_ligolw.types import (
        ToNumPyType as numpytypes,  # noqa: N813
        ToPyType as pytypes,  # noqa: N813
    )

    # if nothing to do...
    if val is None or colname not in cls.validcolumns:
        return val

    llwtype = cls.validcolumns[colname]

    # map to numpy or python types
    try:
        return numpy.sctypeDict[numpytypes[llwtype]](val)
    except KeyError:
        return pytypes[llwtype](val)


# -- identify ------------------------

def is_ligolw(
    origin: Literal["read", "write"],  # noqa: ARG001
    filepath: FileSystemPath | None,  # noqa: ARG001
    fileobj: FileLike | None,
    *args,  # noqa: ANN002
    **kwargs,  # noqa: ARG001
) -> bool:
    """Identify a file object as LIGO_LW-format XML."""
    if fileobj is not None:
        loc = fileobj.tell()
        fileobj.seek(0)
        try:
            line1 = fileobj.readline().lower()
            line2 = fileobj.readline().lower()
            try:
                # binary format
                return (
                    line1.startswith(XML_SIGNATURE)
                    and line2.startswith((LIGOLW_SIGNATURE, LIGOLW_ELEMENT))
                )
            except TypeError:
                # text format
                line1 = cast("str", line1)
                line2 = cast("str", line2)
                return (
                    line1.startswith(XML_SIGNATURE.decode("utf-8"))
                    and line2.startswith((
                        LIGOLW_SIGNATURE.decode("utf-8"),
                        LIGOLW_ELEMENT.decode("utf-8"),
                    ))
                )
        finally:
            fileobj.seek(loc)

    try:
        from igwn_ligolw.ligolw import Element
    except ImportError:
        return False
    return len(args) > 0 and isinstance(args[0], Element)



# Copyright (c) 2019-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Fetch and parse an event catalog from GWOSC."""

from __future__ import annotations

import json
import logging
import numbers
from pathlib import Path
from typing import TYPE_CHECKING

import numpy
from astropy import (
    constants,
    units,
)
from astropy.table import Table
from gwosc.api import (
    DEFAULT_URL as DEFAULT_GWOSC_URL,
    fetch_catalog_json,
)

from ...io.utils import FileSystemPath
from .. import EventTable
from .utils import (
    read_with_columns,
    read_with_where,
)

if TYPE_CHECKING:
    from collections.abc import Iterable

    from ...io.utils import Readable

logger = logging.getLogger(__name__)

#: suffix indicating a unit name
UNIT_SUFFIX = "_unit"

#: custom GWOSC unit mapping
UNITS: dict[str, units.Quantity] = {
    "M_sun X c^2": units.M_sun * constants.c ** 2,
}

#: set of values corresponding to 'missing' or 'null' data
MISSING_DATA: set[str | None] = {
    None,
    "NA",
}

#: values to fill missing data, based on dtype
_FILL_VALUE: dict[type, object] = {
    str: "",
    bytes: b"",
    numbers.Integral: 0,
    numbers.Number: numpy.nan,
}

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _get_unit(
    colname: str,
    unitdict: dict[str, str],
) -> str | units.UnitBase | units.Quantity:
    """Return the unit (name) for a column, or `None` if a match isn't found.

    The return types is either `astropy.units.Unit` or `str`, it doesn't
    matter because setting `column.unit` will automatically convert things
    into `astropy.units.Unit` for us.
    """
    if colname.endswith(("_lower", "_upper")):
        colname = colname.rsplit("_", 1)[0]
    try:
        rawunit = unitdict[colname]
    except KeyError:
        return None
    return UNITS.get(rawunit, rawunit)


def _get_fill_value(type_: type) -> object:
    """Get the fill value for this ``type_``.

    If no default is set for the ``type_``, return `None`.
    """
    for key, replacement in _FILL_VALUE.items():
        if issubclass(type_, key):
            return replacement
    return None


def _mask_replace(
    value: object,
    dtype: type,
) -> object:
    """Replace `value` with the default for the given `dtype`.

    If not default is set for the ``dtype``, just return the
    value unchanged.
    """
    for type_, replacement in _FILL_VALUE.items():
        if issubclass(dtype, type_):
            return replacement
    return value


def _mask_column(col: Iterable) -> tuple[list, list]:
    """Find and replace missing data in a column.

    Returns the new data, and the mask as `list`.
    """
    # find masked indices
    mask = [v in MISSING_DATA for v in col]

    # find common dtype of unmasked values
    dtype = numpy.array([x for i, x in enumerate(col) if not mask[i]]).dtype.type

    # replace the column with a new version that has the masked
    # values replaced by a 'sensible' default for the relevant dtype
    fill_value = _get_fill_value(dtype)
    return (
        [fill_value or x if mask[i] else x for i, x in enumerate(col)],
        mask,
    )


def fetch_catalog(
    catalog: str,
    host: str = DEFAULT_GWOSC_URL,
    **kwargs,
) -> Table:
    """Download a `Table` of events from the GWOSC EventApi."""
    logger.debug("Fetching GWOSC event catalog '%s' from %s", catalog, host)
    # fetch and parse data
    table = parse_eventapi_catalog(
        fetch_catalog_json(catalog, host=host),
        **kwargs,
    )

    # include provenance information in metadata
    table.meta.setdefault("catalog", catalog)
    table.meta.setdefault("host", host)

    return table


def read_catalog(
    source: Readable,
    **kwargs,
) -> Table:
    """Read a `Table` from a GWOSC EventAPI JSON file.

    Parameters
    ----------
    source : `str`, `~pathlib.Path`, `file`
        A file path or file-like object pointing at GWOSC EventAPI
        JSON data.

    kwargs
        Other keyword arguments are passed to `parse_eventapi_catalog`.
    """
    # read a file (object)
    if isinstance(source, FileSystemPath):
        logger.debug("Reading GWOSC event catalog from %s", source)
        with Path(source).open() as file:
            return read_catalog(file)

    # read an open file
    rawdata = json.load(file)

    # and parse it
    return parse_eventapi_catalog(rawdata, **kwargs)


@read_with_columns
@read_with_where
def parse_eventapi_catalog(
    rawdata: dict,
    **kwargs,
) -> Table:
    """Parse a `Table` from the GWOSC EventAPI JSON output.

    Parameters
    ----------
    rawdata : `dict`
        A blob of JSON data from a `gwosc.api.fetch_catalog_json` request.

    kwargs
        All keyword arguments are passed to the
        `~astropy.table.Table` constructor.

    See Also
    --------
    gwosc.api.fetch_catalog_json
        For details of remote data access.

    astropy.table.Table
        For details on how tables are built and what keyword arguments
        are supported.
    """
    data = rawdata["events"]
    coldata: dict[str, list] = {
        "name": list(data.keys()),
    }

    # parse data
    punits = {}
    for event in data.values():
        for key, val in event.items():
            if key.endswith(UNIT_SUFFIX):
                punits[key[:-5]] = val
            else:
                coldata.setdefault(key, []).append(val)

    # rebuild the columns by replacing the masked values
    mask = {}
    for name, col in coldata.items():
        coldata[name], mask[name] = _mask_column(col)

    # convert to columns
    tab = Table(
        coldata,
        masked=True,
        **kwargs,
    )

    # add column metadata
    for name in coldata:
        tab[name].mask = mask[name]
        tab[name].description = name
        tab[name].unit = _get_unit(name, punits)

    # add an index on the event name
    tab.add_index("name")

    logger.debug("Parsed %d events from GWOSC EventAPI", len(tab))
    return tab


EventTable.fetch.registry.register_reader(
    "gwosc",
    EventTable,
    fetch_catalog,
)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.table.io.root`."""

import pytest

from ...segments import (
    Segment,
    SegmentList,
)
from ...testing.utils import assert_table_equal
from .. import filters
from ..filter import filter_table
from .utils import (
    TABLE_CLASSES,
    random_table,
)

# -- test config ---------------------

def pytest_generate_tests(metafunc):
    """Parametrize the ``table`` fixture for each valid table type."""
    if "table" in metafunc.fixturenames:
        metafunc.parametrize("table", TABLE_CLASSES, indirect=True)


@pytest.fixture
def table(request):
    """Create a random table with the right columns."""
    return random_table(
        length=100,
        names=["time", "frequency", "snr"],
        table_class=request.param,
    )


@pytest.fixture
def root_file(table, tmp_path):
    """Write the ``table`` to a (temporary) file and return the path.

    Enables read tests.
    """
    tmp = tmp_path / "table.root"
    table.write(tmp)
    return tmp


# -- tests ---------------------------

@pytest.mark.requires("uproot")
def test_read_write_root(table, tmp_path):
    """Test `Table.read` and `Table.write` with `format='root'`.

    Default options.
    """
    tmp = tmp_path / "table.root"

    # check write
    table.write(tmp)

    # check read gives back same table
    t2 = type(table).read(tmp)
    assert_table_equal(t2, table)


@pytest.mark.requires("uproot")
def test_read_root_where(table, root_file):
    """Test `Table.read(format='root')` with ``where``."""
    # test where works
    segs = SegmentList([Segment(100, 200), Segment(400, 500)])
    t2 = type(table).read(
        root_file,
        where=[
            "200 < frequency < 500",
            ("time", filters.in_segmentlist, segs),
        ],
    )
    assert_table_equal(
        t2,
        filter_table(
            table,
            "frequency > 200",
            "frequency < 500",
            ("time", filters.in_segmentlist, segs),
        ),
    )


@pytest.mark.requires("uproot")
def test_write_root_overwrite(table, tmp_path):
    """Test `Table.write(format='root')` with ``overwrite=True``."""
    tmp = tmp_path / "table.root"
    table.write(tmp)

    # assert failure with overwrite=False (default)
    with pytest.raises(OSError, match="path exists and refusing to overwrite"):
        table.write(tmp)

    # assert works with overwrite=True
    table.write(tmp, overwrite=True)


@pytest.mark.requires("uproot")
def test_write_root_append(table, tmp_path):
    """Test `Table.write(format='root')` with ``append=True``."""
    # write once
    tmp = tmp_path / "table.root"
    table.write(tmp, treename="a")

    # write a second tree
    table.write(tmp, treename="b", append=True)

    # check that we can't read without specifying a tree
    with pytest.raises(
        ValueError,
        match=r"^Multiple trees found",
    ):
        type(table).read(tmp)

    # check that we can read both trees
    type(table).read(tmp, treename="a")
    type(table).read(tmp, treename="b")


@pytest.mark.requires("uproot")
def test_write_root_append_not_found(table, tmp_path):
    """Test `Table.write(format='root', append=True)` when the source doesn't exist."""
    tmp = tmp_path / "table.root"
    with pytest.raises(OSError, match="could not read"):
        table.write(tmp, treename="a", append=True)


@pytest.mark.requires("uproot")
def test_write_root_append_overwrite(table, tmp_path):
    """Test `Table.write(format='root', append=True, overwrite=True)`."""
    # write once
    tmp = tmp_path / "table.root"
    table.write(tmp, treename="a")

    # write a second tree
    table.write(tmp, treename="b", append=True, overwrite=True)

    # check that we can read the new tree
    type(table).read(tmp, treename="b")

    # if overwriting, check that we can't read the old tree any more
    with pytest.raises(KeyError):
        type(table).read(tmp, treename="a")



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Basic HDF5 I/O methods for Array and sub-classes."""

from __future__ import annotations

import contextlib
import pickle
from decimal import Decimal
from functools import (
    partial,
    wraps,
)
from operator import attrgetter
from typing import TYPE_CHECKING

from astropy.units import (
    Quantity,
    UnitBase,
)

from ...detector import Channel
from ...io import hdf5 as io_hdf5
from ...time import LIGOTimeGPS
from .. import (
    Array,
    Index,
    Series,
)

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path
    from typing import (
        IO,
        Any,
    )

    import h5py

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"

ATTR_TYPE_MAP: dict[type, Callable] = {
    Quantity: attrgetter("value"),
    Channel: str,
    UnitBase: str,
    Decimal: float,
    LIGOTimeGPS: float,
}


# -- read ----------------------------


@io_hdf5.with_read_hdf5
def read_hdf5_array(
    source: h5py.Group,
    path: str | None = None,
    array_type: type[Array] = Array,
) -> Array:
    """Read an `Array` from the given HDF5 object.

    Parameters
    ----------
    source : `str`, `Path`, `file`, `h5py.File`, `h5py.Group`
        Path to HDF5 file on disk, or open `h5py.Group`.

    path : `str`, optional
        Path of dataset in HDF5 file.
        This option is required if the ``source`` contains more than a single
        dataset.

    array_type : `type`
        Desired return type.
    """
    dataset = io_hdf5.find_dataset(source, path=path)
    attrs = dict(dataset.attrs)

    # unpickle channel object
    with contextlib.suppress(KeyError):
        attrs["channel"] = _unpickle_channel(attrs["channel"])

    # unpack byte strings for python3
    for key, attr in attrs.items():
        if isinstance(attr, bytes):
            attrs[key] = attr.decode("utf-8")
    return array_type(dataset[()], **attrs)


def _unpickle_channel(raw: str | bytes, **kwargs) -> str:
    """Try and unpickle a channel with sensible error handling."""
    if isinstance(raw, str):
        return raw

    try:
        return pickle.loads(raw, **kwargs)
    except (
        ValueError,
        pickle.UnpicklingError,
        EOFError,
        IndexError,
    ):
        raw = raw.decode("utf-8")
        return _unpickle_channel(raw)


# -- write ---------------------------

class IgnoredAttributeError(ValueError):
    """Internal exception to indicate an attribute to be ignored."""


def _format_metadata_attribute(value: Any) -> Any:
    """Format a value for writing to HDF5 as a `h5py.Dataset` attribute."""
    if value is None or (isinstance(value, Index) and value.regular):
        raise IgnoredAttributeError

    # map type to something HDF5 can handle
    for typekey, func in ATTR_TYPE_MAP.items():
        if issubclass(type(value), typekey):
            return func(value)
    return value


def write_array_metadata(
    dataset: h5py.Dataset,
    array: Array,
) -> None:
    """Write metadata for ``array`` into the `h5py.Dataset`."""
    for attr in ("unit", *array._metadata_slots):
        # format attribute
        try:
            value = _format_metadata_attribute(
                getattr(array, f"_{attr}", None),
            )
        except IgnoredAttributeError:
            continue

        # store attribute
        try:
            dataset.attrs[attr] = value
        except (
            TypeError,
            ValueError,
            RuntimeError,
        ) as exc:
            exc.args = (
                f"Failed to store {attr} ({type(value).__name__}) "
                f"for {type(array).__name__}: '{exc}'",
            )
            raise


@io_hdf5.with_write_hdf5
def write_hdf5_array(
    array: Array,
    h5g: h5py.Group,
    path: str | None = None,
    *,
    attrs: dict[str, Any] | None = None,
    append: bool = False,  # used by `with_write_hdf5`  # noqa: ARG001
    overwrite: bool = False,
    compression: str = "gzip",
    **kwargs,
) -> h5py.Dataset:
    """Write ``array`` to a `h5py.Dataset`.

    Parameters
    ----------
    array : `gwpy.types.Array`
        The data object to write.

    h5g : `str`, `Path`, `file`, `h5py.Group`
        A file path to write to, or an `h5py.Group` in which to create
        a new dataset.

    path : `str`, optional
        The path inside the group at which to create the new dataset,
        defaults to ``array.name``.

    attrs : `dict`, optional
        Extra metadata to write into `h5py.Dataset.attrs`, on top of
        the default metadata.

    append : `bool`, default: `False`
        Tf `True`, write new dataset to existing file, otherwise an
        exception will be raised if the output file exists.
        Default is `False`.

    overwrite : `bool`, default: `False`
        If `True`, overwrite an existing dataset in an existing file,
        otherwise an exception will be raised if a dataset exists with
        the given name.
        Default is `False`.

    compression : `str`, `int`, optional
        Compression option to pass to :meth:`h5py.Group.create_dataset`.

    kwargs
        Other keyword arguments for :meth:`h5py.Group.create_dataset`.

    Returns
    -------
    datasets : `h5py.Dataset`
        The newly created dataset.

    See Also
    --------
    h5py.Group.create_dataset
        For documentation of other valid keyword arguments.
    """
    if path is None:
        path = array.name
    if path is None:
        msg = (
            f"Cannot determine HDF5 path for {type(array).__name__}, "
            "please set `name` attribute, or pass `path` keyword when writing"
        )
        raise ValueError(
            msg,
        )

    # create dataset
    dset = io_hdf5.create_dataset(
        h5g,
        path,
        overwrite=overwrite,
        data=array.value,
        compression=compression,
        **kwargs,
    )

    # write default metadata
    write_array_metadata(dset, array)

    # allow caller to specify their own metadata dict
    if attrs:
        for key in attrs:
            dset.attrs[key] = attrs[key]

    return dset


def format_index_array_attrs(series: Series) -> dict[str, Any]:
    """Format metadata attributes for and indexed array.

    This function is used to provide the necessary metadata to meet
    the (proposed) LIGO Common Data Format specification for series data
    in HDF5.
    """
    attrs = {}
    # loop through named axes
    for _, axis in zip(range(series.ndim), ("x", "y", "z"), strict=False):
        # find property names
        unit = f"{axis}unit"
        origin = f"{axis}0"
        delta = f"d{axis}"

        # store attributes
        aunit = getattr(series, unit)
        attrs.update(
            {
                unit: str(aunit),
                origin: getattr(series, origin).to(aunit).value,
                delta: getattr(series, delta).to(aunit).value,
            },
        )
    return attrs


@wraps(write_hdf5_array)
def write_hdf5_series(
    series: Series,
    output: str | Path | IO | h5py.Group,
    path: str | None = None,
    attrs: dict[str, Any] | None = None,
    **kwargs,
) -> h5py.Dataset:
    """Write a Series to HDF5.

    See `write_hdf5_array` for details of arguments and keywords.
    """
    if attrs is None:
        attrs = format_index_array_attrs(series)
    return write_hdf5_array(
        series,
        output,
        path=path,
        attrs=attrs,
        **kwargs,
    )


# -- register ------------------------

def register_hdf5_array_io(
    array_type: type[Array],
    format: str = "hdf5",  # noqa: A002
    *,
    read: bool = True,
    write: bool = True,
    identify: bool = True,
) -> None:
    """Registry read() and write() methods for the HDF5 format."""
    if read:
        array_type.read.registry.register_reader(
            format,
            array_type,
            partial(read_hdf5_array, array_type=array_type),
        )

    if write:
        if issubclass(array_type, Series):
            writer = write_hdf5_series
        else:
            writer = write_hdf5_array
        array_type.write.registry.register_writer(
            format,
            array_type,
            writer,
        )

    if identify:
        array_type.read.registry.register_identifier(
            format,
            array_type,
            io_hdf5.identify_hdf5,
        )



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilties for the GWpy test suite."""

# ruff: noqa: PT028, S101

from __future__ import annotations

import subprocess
import tempfile
from itertools import zip_longest
from pathlib import Path
from shutil import which
from typing import TYPE_CHECKING

import numpy
import pytest
from astropy.time import Time
from numpy.testing import (
    assert_allclose,
    assert_array_equal,
)

from ..io.cache import file_segment

if TYPE_CHECKING:
    from collections.abc import (
        Callable,
        Container,
        Iterable,
    )
    from typing import Any

    from astropy.units import Quantity

    from ..segments import (
        DataQualityFlag,
        Segment,
    )
    from ..signal.filter_design import ZpkType
    from ..types import Array

KLIST = which("klist")

# -- useful constants ----------------

TEST_DATA_PATH = Path(__file__).parent / "data"
TEST_DATA_DIR = str(TEST_DATA_PATH)
TEST_GWF_FILE = str(TEST_DATA_PATH / "HLV-HW100916-968654552-1.gwf")
TEST_GWF_SPAN = file_segment(TEST_GWF_FILE)
TEST_HDF5_FILE = str(TEST_DATA_PATH / "HLV-HW100916-968654552-1.hdf")


# -- dependencies --------------------

def _has_kerberos_credential() -> bool:
    """Return `True` if the current user has a valid kerberos credential.

    This function just calls ``klist -s`` and returns `True` if the
    command returns a zero exit code, and `False` if it doesn't, or
    the call fails in any other way.
    """
    if KLIST is None:
        return False
    try:
        subprocess.check_call(
            [KLIST, "-s"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
    except (
        subprocess.CalledProcessError,  # klist failed (no credential)
        FileNotFoundError,  # klist isn't there
    ):
        return False
    return True


#: skip a test function if no kerberos credential is found
skip_kerberos_credential = pytest.mark.skipif(
    not _has_kerberos_credential(),
    reason="no kerberos credential",
)


# -- assertions ----------------------

def assert_quantity_equal(
    q1: Quantity,
    q2: Quantity,
) -> None:
    """Assert that two `~astropy.units.Quantity` objects are the same."""
    _assert_quantity(
        q1,
        q2,
        array_assertion=assert_array_equal,
    )


def assert_quantity_almost_equal(
    q1: Quantity,
    q2: Quantity,
) -> None:
    """Assert that two `~astropy.units.Quantity` objects are almost the same.

    This method asserts that the units are the same and that the values are
    equal within precision.
    """
    _assert_quantity(
        q1,
        q2,
        array_assertion=assert_allclose,
    )


def _assert_quantity(
    q1: Quantity,
    q2: Quantity,
    array_assertion: Callable = assert_array_equal,
) -> None:
    assert q1.unit == q2.unit, f"'{q1.unit}' != '{q2.unit}'"
    array_assertion(q1.value, q2.value)


def assert_quantity_sub_equal(
    a: Array,
    b: Array,
    *attrs: str,
    almost_equal: bool = False,
    exclude: Container[str] | None = None,
    **kwargs,
) -> None:
    """Assert that two `~gwpy.types.Array` objects are the same (or almost).

    Parameters
    ----------
    a, b : `~gwpy.types.Array`
        The arrays to be tested (can be subclasses).

    *attrs
        The list of attributes to test, defaults to all.

    almost_equal : `bool`, optional
        Allow the numpy array's to be 'almost' equal, default: `False`,
        i.e. require exact matches.

    exclude : `list`, optional
        A list of attributes to exclude from the test.

    kwargs
        Other keyword arguments are passed to the array comparison operator
        `numpy.testing.assert_array_equal` or `numpy.testing.assert_allclose`.

    See Also
    --------
    numpy.testing.assert_array_equal
    numpy.testing.assert_allclose
    """
    # get value test method
    assert_array: Callable
    if almost_equal:
        assert_array = assert_allclose
    else:
        assert_array = assert_array_equal

    # parse attributes to be tested
    if not attrs:
        attrs = a._metadata_slots
    checkattrs = [attr for attr in attrs if attr not in (exclude or [])]

    # don't assert indexes that don't exist for both
    def _check_index(dim: str) -> None:
        index = f"{dim}index"
        _index = "_" + index
        if (
            index in attrs
            and getattr(a, _index, "-") == "-"
            and getattr(b, _index, "-") == "-"
        ):
            checkattrs.remove(index)
    _check_index("x")
    _check_index("y")

    # test data
    assert_attributes(a, b, *checkattrs)
    assert_array(a.value, b.value, **kwargs)


def assert_attributes(
    a: Array,
    b: Array,
    *attrs: str,
) -> None:
    """Assert that the attributes for two objects match.

    `attrs` should be `list` of attribute names that can be accessed
    with `getattr`
    """
    for attr in attrs:
        x = getattr(a, attr, None)
        y = getattr(b, attr, None)
        if isinstance(x, numpy.ndarray) and isinstance(y, numpy.ndarray):
            assert_array_equal(x, y)
        elif isinstance(x, Time) and isinstance(y, Time):
            assert x.gps == y.gps
        else:
            assert x == y


def assert_table_equal(
    a: Array,
    b: Array,
    *,
    is_copy: bool = True,
    meta: bool = False,
    check_types: bool = True,
    almost_equal: bool = False,
) -> None:
    """Assert that two tables store the same information."""
    # check column names are the same
    assert sorted(a.colnames) == sorted(b.colnames)

    # check that the metadata match
    if meta:
        assert a.meta == b.meta

    assert_array: Callable
    if almost_equal:
        assert_array = assert_allclose
    else:
        assert_array = assert_array_equal

    # actually check the data
    for name in a.colnames:
        cola = a[name]
        colb = b[name]
        if check_types:
            assert cola.dtype == colb.dtype, \
                f"{name} dtype mismatch: {cola.dtype} != {colb.dtype}"
        assert_array(cola, colb)

    # check that the tables are copied or the same data
    for name in a.colnames:
        # check may_share_memory is True when copy is False and so on
        assert numpy.may_share_memory(a[name], b[name]) is not is_copy


def assert_segmentlist_equal(
    a: Iterable[Segment],
    b: Iterable[Segment],
) -> None:
    """Assert that two `SegmentList`s contain the same data."""
    for aseg, bseg in zip_longest(a, b):
        assert aseg == bseg


def assert_flag_equal(
    a: DataQualityFlag,
    b: DataQualityFlag,
    attrs: Iterable[str] = ("name", "ifo", "tag", "version"),
) -> None:
    """Assert that two `DataQualityFlag`s contain the same data."""
    assert_segmentlist_equal(a.active, b.active)
    assert_segmentlist_equal(a.known, b.known)
    for attr in attrs:
        assert getattr(a, attr) == getattr(b, attr)


def assert_dict_equal(
    a: dict[Any, Any],
    b: dict[Any, Any],
    assert_value: Callable,
    **kwargs,
) -> None:
    """Assert that two `dict`s contain the same data.

    Parameters
    ----------
    a, b
        Two objects to compare.

    assert_value : `callable`
        Method to compare that two dict entries are the same.

    kwargs
        Keyword arguments to pass to ``assert_value``.
    """
    assert a.keys() == b.keys()
    for key in a | b:
        assert_value(a[key], b[key], **kwargs)


def assert_zpk_equal(
    a: ZpkType,
    b: ZpkType,
    *,
    almost_equal: bool = False,
) -> None:
    """Assert that two ZPK filters are the same (or almost)."""
    assert_array: Callable
    if almost_equal:
        assert_array = assert_allclose
    else:
        assert_array = assert_array_equal
    z1, p1, k1 = a
    z2, p2, k2 = b
    assert_array(z1, z2)  # zeros
    assert_array(p1, p2)  # poles
    assert_array(k1, k2)  # gain


# -- I/O helpers ---------------------

def test_read_write(
    data: Array,
    fmt: str,
    extension: str | None = None,
    *,
    autoidentify: bool = True,
    read_args: Iterable[Any] = [],
    read_kw: dict[str, Any] | None = None,
    write_args: Iterable[Any] = [],
    write_kw: dict[str, Any] | None = None,
    assert_equal: Callable = assert_quantity_sub_equal,
    assert_kw: dict[str, Any] | None = None,
) -> None:
    """Test that data can be written to and read from a file in some format.

    Parameters
    ----------
    data : some type with `.read()` and `.write()` methods
        the data to be written

    fmt : `str`
        the name of the file format (as registered with `astropy.io.registry`

    extension : `str`, optional
        the name of the file extension, defaults to ``.<fmt>``

    autoidentify : `bool`, optional
        attempt to auto-identify when reading writing by not specifying
        ``format``

    read_args : `list`, optional
        positional arguments to pass to ``type(data).read()``

    read_kw : `dict`, optional
        keyword arguments to pass to ``type(data).read()``

    write_args : `list`, optional
        positional arguments to pass to ``data.write()``

    write_kw : `dict`, optional
        keyword arguments to pass to ``data.write()``

    assert_equal : `callable`, optional
        the function to assert that the object read back from file matches
        the original ``data``

    assert_kw : `dict`, optional
        keyword arguments to pass to ``assert_equal``
    """
    # parse extension and add leading period
    if assert_kw is None:
        assert_kw = {}
    if write_kw is None:
        write_kw = {}
    if read_kw is None:
        read_kw = {}
    if extension is None:
        extension = fmt
    extension = extension.lstrip(".")

    DataClass = type(data)  # noqa: N806

    with tempfile.TemporaryDirectory() as tmpdir:
        tmp = Path(tmpdir) / f"test.{extension}"

        data.write(tmp, *write_args, format=fmt, **write_kw)

        # try again with automatic format identification
        if autoidentify:
            data.write(str(tmp), *write_args, **write_kw)

        # read the data back and check that its the same
        new = DataClass.read(tmp, *read_args, format=fmt, **read_kw)
        assert_equal(new, data, **assert_kw)

        # try again with automatic format identification
        if autoidentify:
            new = DataClass.read(str(tmp), *read_args, **read_kw)
            assert_equal(new, data, **assert_kw)



# Copyright (c) 2014-2017 Louisiana State University
#               2017-2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import os
import re
import warnings
from pathlib import Path

import matplotlib
import sphinx_github_style
from numpydoc import docscrape_sphinx
from sphinx.util import logging

import gwpy
from gwpy.utils.sphinx import gallery as cli_gallery

TODAY = datetime.datetime.now(tz=datetime.UTC).date()

SPHINX_DIR = Path(__file__).parent.absolute()
STATIC_DIRNAME = "_static"

PROJECT_URL = "https://gitlab.com/gwpy/gwpy"

# use caching in GWpy calls
os.environ.update({
    "GWPY_CACHE": "true",
})

# -- versions ------------------------

GWPY_VERSION = gwpy.__version__

# parse version number to get git reference
_setuptools_scm_version_regex = re.compile(
    r"\+g(\w+)(?:\Z|\.)",
)
if match := _setuptools_scm_version_regex.search(GWPY_VERSION):
    GWPY_GIT_REF, = match.groups()
else:
    GWPY_GIT_REF = f"v{GWPY_VERSION}"

# -- matplotlib ----------------------

matplotlib.use("agg")

# ignore warnings that aren't useful for documentation
if "CI" not in os.environ:
    for message, category in (
        (".*non-GUI backend.*", UserWarning),
        (".*gwpy.plot.*", DeprecationWarning),
        ("", matplotlib.MatplotlibDeprecationWarning),
    ):
        warnings.filterwarnings("ignore", message=message, category=category)

# -- general -------------------------

needs_sphinx = "4.0"
project = "GWpy"
copyright = f"{TODAY.year} Cardiff University"
version = "dev" if ".dev" in GWPY_VERSION else GWPY_VERSION
release = GWPY_VERSION

# extension modules
# DEVNOTE: please make sure and add 3rd-party dependencies to
#          pyproject.toml's [project.optional-dependencies]/docs
extensions = [
    "matplotlib.sphinxext.plot_directive",
    "matplotlib.sphinxext.roles",
    "myst_parser",
    "numpydoc",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.ifconfig",
    "sphinx.ext.inheritance_diagram",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.mathjax",
    "sphinx.ext.todo",
    "sphinx_automodapi.automodapi",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinx_gallery.gen_gallery",
    "sphinxarg.ext",
    "sphinxcontrib.programoutput",
]

# content management
default_role = "obj"
exclude_patterns = [
    "_references.rst",
    "_build",
    "_generated",
    "cli/_examples*",
]
templates_path = [
    "_templates",
]

# epilog
rst_epilog = "\n.. include:: /_references.rst"

# -- HTML formatting -----------------

html_theme = "pydata_sphinx_theme"

html_title = f"{project} {version}"
html_logo = "gwpy-logo.png"

html_static_path = [STATIC_DIRNAME]
html_favicon = str(Path(STATIC_DIRNAME) / "favicon.png")
html_copy_source = False

html_theme_options = {
    "gitlab_url": PROJECT_URL,
    "pygments_light_style": "default",
    "pygments_dark_style": "material",
    "use_edit_page_button": True,
}
html_context = {
    "gitlab_user": "gwpy",
    "gitlab_repo": "gwpy",
    "gitlab_version": "main",
    "doc_path": "docs",
}

html_css_files = [
    # GWpy customisations
    "css/gwpy-sphinx.css",
]

# Fold long signature lines
maximum_signature_line_length = 80

# -- extensions config ---------------

# -- autodoc

autoclass_content = "class"
autodoc_default_flags = ["show-inheritance", "members", "inherited-members"]

# -- automodapi

automodapi_toctreedirnm = "reference"
automodapi_writereprocessed = False

# -- autosummary

autosummary_generate = True

# -- copybutton

copybutton_prompt_text = " |".join((  # noqa: FLY002
    ">>>",
    r"\.\.\.",
    r"\$"
    r"In \[\d*\]:",
    r" {2,5}\.\.\.:",
    " {5,8}: ",
))
copybutton_prompt_is_regexp = True
copybutton_line_continuation_character = "\\"

# -- extlinks

extlinks = {
    # GWpy project links
    "issue": (f"{PROJECT_URL}/-/issues/%s", "gwpy/gwpy#%s"),
    "mr": (f"{PROJECT_URL}/-/merge_requests/%s", "gwpy/gwpy!%s"),
    "release": (f"{PROJECT_URL}/-/releases/%s", "%s"),
    # DCC document
    "dcc": ("https://dcc.ligo.org/%s/public", "%s"),
    # DOI link
    "doi": ("https://doi.org/%s", "doi:%s"),
    # GitHub user profile
    "ghu": ("https://github.com/%s", "@%s"),
    # GitLab User profile
    "glu": ("https://gitlab.com/%s", "@%s"),
}

# -- inhertance_diagram

# configure inheritance diagram
inheritance_graph_attrs = {"rankdir": "TB"}

# -- intersphinx

# Intersphinx
intersphinx_mapping = {key: (value, None) for key, value in {
    "astropy": "https://docs.astropy.org/en/stable/",
    "coloredlogs": "https://coloredlogs.readthedocs.io/en/latest/",
    "dateparser": "https://dateparser.readthedocs.io/en/stable/",
    "dateutil": "https://dateutil.readthedocs.io/en/stable/",
    "dqsegdb2": "https://dqsegdb2.readthedocs.io/en/stable/",
    "gssapi": "https://pythongssapi.github.io/python-gssapi/",
    "gwdatafind": "https://gwdatafind.readthedocs.io/en/stable/",
    "gwosc": "https://gwosc.readthedocs.io/en/stable/",
    "h5py": "https://docs.h5py.org/en/latest/",
    "humanfriendly": "https://humanfriendly.readthedocs.io/en/latest/",
    "igwn-auth-utils": "https://igwn-auth-utils.readthedocs.io/en/stable/",
    "igwn-ligolw": "https://igwn-ligolw.readthedocs.io/en/stable/",
    "igwn-segments": "https://igwn-segments.readthedocs.io/en/stable/",
    "ligo.skymap": "https://lscsoft.docs.ligo.org/ligo.skymap/",
    "ligotimegps": "https://ligotimegps.readthedocs.io/en/stable/",
    "lscsoft-glue": "https://lscsoft.docs.ligo.org/glue/",
    "matplotlib": "https://matplotlib.org/",
    "numpy": "https://numpy.org/doc/stable/",
    "pycbc": "https://pycbc.org/pycbc/latest/html/",
    "python": "https://docs.python.org/3/",
    "requests-pelican": "https://requests-pelican.readthedocs.io/en/stable/",
    "requests-scitokens": "https://requests-scitokens.readthedocs.io/en/stable/",
    "uproot": "https://uproot.readthedocs.io/en/stable/",
    "scipy": "https://docs.scipy.org/doc/scipy/reference/",
    "scitokens": "https://scitokens.readthedocs.io/en/stable/",
}.items()}
intersphinx_resolve_self = "gwpy"

# -- linkcode

# linkcode
linkcode_url = sphinx_github_style.get_linkcode_url(
    blob=sphinx_github_style.get_linkcode_revision(
        os.getenv("READTHEDOCS_GIT_COMMIT_HASH") or "head",
    ),
    url=PROJECT_URL,
).replace("/blob/", "/-/blob/")
linkcode_resolve = sphinx_github_style.get_linkcode_resolve(
    linkcode_url,
    repo_dir=os.getenv("READTHEDOCS_REPOSITORY_PATH"),
)

# -- myst_parser

myst_enable_extensions = [
    "attrs_block",
]

# -- numpydoc

# fix numpydoc autosummary
numpydoc_show_class_members = False

# use blockquotes (numpydoc>=0.8 only)
numpydoc_use_blockquotes = True

# auto-insert plot directive in examples
numpydoc_use_plots = True

# enable cross-referencing of parameter types
numpydoc_xref_param_type = True
numpydoc_xref_ignore = {
    "of",
    "optional",
    "or",
    "subclass",
}

# update the plot detection to include .show() calls
parts = re.split(r"[\(\)|]", docscrape_sphinx.IMPORT_MATPLOTLIB_RE)[1:-1]
parts.extend(("fig.show()", "plot.show()"))
docscrape_sphinx.IMPORT_MATPLOTLIB_RE = r"\b({})\b".format("|".join(parts))

# -- plot_directive

plot_rcparams = dict(matplotlib.rcParams)
plot_rcparams.update({
    "backend": "agg",
})
plot_apply_rcparams = True
plot_formats = ["png"]
plot_include_source = True
plot_html_show_source_link = False

# -- sphinx-gallery

sphinx_gallery_conf = {
    # where to find the examples
    "examples_dirs": [
        str(Path("..") / "examples"),
        str(Path() / "cli" / "_examples"),
    ],
    # where to render them
    "gallery_dirs": [
        str(Path() / "examples"),
        str(Path() / "cli" / "examples"),
    ],
    "download_all_examples": False,
    "filename_pattern": r"/.*\.py",  # execute all examples
    "ignore_pattern": r"test_.*\.py",  # ignore example tests
    "reset_modules": [],  # don't reset matplotlib (or seaborn)
    "write_computation_times": False,
    "within_subsection_order": "ExampleTitleSortKey",
    # allow some examples to fail
    "only_warn_on_example_error": True,
}

# -- build CLI examples --------------

def render_cli_examples(app):
    """Render the CLI examples for sphinx-gallery."""
    clidir = Path(app.confdir) / "cli"
    return cli_gallery.render_entry_point_examples(
        clidir / "examples.ini",
        clidir / "_examples",
        logger=logging.getLogger("gwpy-plot-gallery"),
        filename_prefix="gwpy-plot-",
    )


# -- setup sphinx --------------------

def setup(app):
    # render the CLI examples, with low priority number so that this occurs
    # before sphinx-gallery attempts to execute them
    app.connect("builder-inited", render_cli_examples, priority=10)



# Copyright (c) 2025 Cardiff University
#
# This file is part of GWpy.
#
# GWpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# GWpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with GWpy.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for :mod:`gwpy.io.pelican`."""

from unittest import mock

import pytest
from urllib3.util import parse_url

from ...testing.errors import pytest_skip_flaky_network
from .. import pelican as io_pelican

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"


def _protocols():
    return ["pelican", "osdf"]


@mock.patch("gwpy.io.pelican._pelican_protocols", _protocols)
@pytest.mark.parametrize(("url", "result"), [
    pytest.param("/path/to/data", False, id="path"),
    pytest.param("https://example.com/path/to/data", False, id="http"),
    pytest.param("pelican://example.com/path/to/data", True, id="pelican"),
    pytest.param("osdf:///path/to/data", True, id="osdf"),
])
def test_is_pelican_url(url, result):
    """Test `is_pelican_url()`."""
    assert io_pelican.is_pelican_url(url) is result


@pytest_skip_flaky_network
@pytest.mark.requires("requests_pelican")
def test_query_director():
    """Test `query_director()`."""
    urls, needauth, authkw = io_pelican.query_director(
        "osdf:///igwn/ligo",
    )
    assert len(urls) >= 1
    for u in urls:
        parsed = parse_url(u)
        assert parsed.path == "/igwn/ligo"
    assert needauth is True
    # /igwn namespace doesn't get issuer info
    assert "issuer" not in authkw


@pytest_skip_flaky_network
@pytest.mark.requires("requests_pelican")
def test_download_file():
    """Test `download_file()`."""
    path = io_pelican.download_file(
        "osdf:///gwdata/zenodo/README.zenodo",
        cache=False,
    )
    with open(path) as file:
        assert next(file).strip() == "## Mirror of IGWN Zenodo Communities"


@pytest_skip_flaky_network
@pytest.mark.requires("requests_pelican")
def test_open_remote_file():
    """Test `open_remote_file()`."""
    with io_pelican.open_remote_file(
        "osdf:///gwdata/zenodo/README.zenodo",
        cache=False,
    ) as file:
        assert next(file).strip() == "## Mirror of IGWN Zenodo Communities"
